from typing import TypeVar, Type, Callable, Any, Tuple, Generic
from concurrent.futures import wait, ThreadPoolExecutor

from neovim.msgpack_rpc.event_loop.base import BaseEventLoop

import amino
from amino import _, L, __, Try, IO, Lists, Either, List
from amino.do import do, Do
from amino.dispatch import dispatch_alg
from amino.util.exception import format_exception
from amino.algebra import Algebra
from amino.io import IOException

from ribosome.nvim import NvimFacade, NvimIO
from ribosome.logging import ribo_log, Logging
from ribosome import NvimPlugin
from ribosome.config import Config
from ribosome.nvim.io import NS
from ribosome.dispatch.run import DispatchJob, RunDispatchSync, RunDispatchAsync, invalid_dispatch, execute_data_trans
from ribosome.dispatch.data import (DispatchError, DispatchReturn, DispatchUnit, DispatchOutput, DispatchSync,
                                    DispatchAsync, DispatchResult, Dispatch, DispatchIO, IODIO, DIO, DispatchErrors,
                                    NvimIODIO, DispatchOutputAggregate, GatherIOsDIO, DispatchDo, GatherSubprocsDIO)
from ribosome.plugin_state import PluginState, PluginStateHolder
from ribosome.trans.queue import PrioQueue
from ribosome.trans.message_base import Message
from ribosome.dispatch.loop import process_message
from ribosome.trans.send_message import send_message
from ribosome.dispatch.transform import validate_trans_action
from ribosome.trans.action import TransM, TransMPure, TransMBind
from ribosome.trans.handler import TransComplete

Loop = TypeVar('Loop', bound=BaseEventLoop)
NP = TypeVar('NP', bound=NvimPlugin)
D = TypeVar('D')
AS = TypeVar('AS')
A = TypeVar('A', bound=AS)
B = TypeVar('B')
C = TypeVar('C', bound=Config)
R = TypeVar('R')
DP = TypeVar('DP', bound=Dispatch)
RDP = TypeVar('RDP', bound=Algebra)
Res = NS[PluginState[D], DispatchResult]


def gather_ios(ios: List[IO[A]], timeout: float) -> List[Either[IOException, A]]:
    with ThreadPoolExecutor(thread_name_prefix='ribosome_dio') as executor:
        futures = ios.map(lambda i: executor.submit(i.attempt_run))
        completed, timed_out = wait(futures, timeout=timeout)
        return Lists.wrap(completed).map(__.result(timeout=timeout))


class ExecuteDispatchIO(Logging):

    def iodio(self, io: IODIO[A]) -> NS[PluginState[D], TransComplete]:
        return NS.from_io(io.io)

    def gather_i_os_dio(self, io: GatherIOsDIO[A]) -> NS[PluginState[D], TransComplete]:
        def gather() -> R:
            gio = io.io
            return gio.handle_result(gather_ios(gio.ios, gio.timeout))
        return NS.from_io(IO.delay(gather))

    def gather_subprocs_dio(self, io: GatherSubprocsDIO[A, TransComplete]) -> NS[PluginState[D], TransComplete]:
        def gather() -> TransComplete:
            gio = io.io
            popens = gio.procs.map(__.execute(gio.timeout))
            return gio.handle_result(gather_ios(popens, gio.timeout))
        return NS.from_io(IO.delay(gather))

    def nvim_iodio(self, io: NvimIODIO[A]) -> NS[PluginState[D], TransComplete]:
        return NS.lift(io.io)


execute_io = dispatch_alg(ExecuteDispatchIO(), DIO, '')


@do(NS[PluginState[D], R])
def run_trans_m(tr: TransM) -> Do:
    if isinstance(tr, TransMPure):
        result = yield execute_data_trans(tr.handler)
        yield normalize_output(result)
    elif isinstance(tr, TransMBind):
        result = yield run_trans_m(tr.fa)
        n = tr.f(result)
        yield run_trans_m(n)


class ExecuteDispatchOutput(Logging):

    def dispatch_error(self, result: DispatchError) -> NS[PluginState[D], R]:
        io = result.exception / NvimIO.exception | NvimIO.delay(lambda v: ribo_log.error(result.message))
        return NS.lift(io)

    def dispatch_errors(self, result: DispatchErrors) -> NS[PluginState[D], R]:
        return result.errors.traverse(self.dispatch_error, NS)

    def dispatch_return(self, result: DispatchReturn) -> NS[PluginState[D], R]:
        return NS.pure(result.value)

    def dispatch_unit(self, result: DispatchUnit) -> NS[PluginState[D], R]:
        return NS.pure(0)

    @do(NS[PluginState[D], R])
    def dispatch_io(self, result: DispatchIO) -> Do:
        custom_executor = yield NS.inspect(_.dispatch_config.io_executor)
        executor = custom_executor | (lambda: execute_io)
        inner = yield executor(result.io)
        result = yield validate_trans_action(TransComplete('io', inner))
        yield normalize_output(result)

    @do(NS[PluginState[D], R])
    def dispatch_output_aggregate(self, result: DispatchOutputAggregate) -> Do:
        yield result.results.traverse(normalize_output, NS)
        yield DispatchResult.unit_nio

    def dispatch_do(self, result: DispatchDo) -> NS[PluginState[D], R]:
        return run_trans_m(result.trans.action)


execute_output = dispatch_alg(ExecuteDispatchOutput(), DispatchOutput, '')


@do(NS[PluginState[D], R])
def normalize_output(result: DispatchResult) -> Do:
    yield NS.modify(__.enqueue(result.msgs))
    yield execute_output(result.output)


@do(NS[PluginState[D], R])
def run_dispatch(action: Callable[[], NS[PluginState[D], B]], unpack: Callable[[B], Res]) -> Do:
    response = yield action()
    result = yield unpack(response)
    yield normalize_output(result)


# FIXME check how long-running messages can be handled; timeout for acquire is 10s
def exclusive(holder: PluginStateHolder, f: Callable[[], NvimIO[R]], desc: str) -> NvimIO[R]:
    def release(error: Any) -> None:
        holder.release()
    yield NvimIO.delay(lambda v: holder.acquire())
    ribo_log.debug(f'exclusive: {desc}')
    state, response = yield f().error_effect(release)
    yield NvimIO.delay(lambda v: holder.update(state))
    yield NvimIO.delay(release)
    ribo_log.debug(f'release: {desc}')
    yield NvimIO.pure(response)


@do(NvimIO[R])
def exclusive_dispatch(holder: PluginStateHolder,
                       action: Callable[[], NS[PluginState[D], B]],
                       unpack: Callable[[B], NS[PluginState[D], DispatchResult]],
                       desc: str) -> Do:
    return exclusive(holder, lambda: run_dispatch(action, unpack).run(holder.state), desc)


class DispatchRunner(Generic[RDP]):

    @staticmethod
    def cons(run: Type[RDP], dp: Type[DP]) -> 'DispatchRunner[RDP]':
        return DispatchRunner(lambda args: dispatch_alg(run(args), dp, '', invalid_dispatch))

    def __init__(self, f: Callable[[tuple], Callable[[RDP], NS[PluginState[D], DispatchOutput]]]) -> None:
        self.f = f

    def __call__(self, args: tuple) -> Callable[[RDP], NS[PluginState[D], DispatchOutput]]:
        return self.f(args)


sync_runner = DispatchRunner.cons(RunDispatchSync, DispatchSync)
async_runner = DispatchRunner.cons(RunDispatchAsync, DispatchAsync)


def sync_sender(job: DispatchJob, dispatch: DP, runner: DispatchRunner[RDP]) -> Callable[[], Res]:
    def send() -> NS[PluginState[D], DispatchResult]:
        return runner(job.args)(dispatch)
    return send


def execute(job: DispatchJob, dispatch: DP, runner: DispatchRunner[RDP]) -> NvimIO[Any]:
    return exclusive_dispatch(job.state, sync_sender(job, dispatch, runner), NS.pure, dispatch.desc)


def fork_job(f: Callable[[], NvimIO[None]], job: DispatchJob) -> NvimIO[None]:
    def run(vim: NvimFacade) -> None:
        (Try(f) // __.attempt(vim)).leffect(L(request_error)(job, _))
        ribo_log.debug(f'async job {job.name} completed')
    return NvimIO.fork(run)


@do(NvimIO)
def execute_sync(job: DispatchJob, dispatch: DispatchSync) -> Do:
    response = yield execute(job, dispatch, sync_runner)
    yield fork_job(lambda: resend_loop(job.state), job)
    yield NvimIO.pure(response)


def request_error(job: DispatchJob, exc: Exception) -> int:
    sync_prefix = '' if job.sync else 'a'
    desc = f'{sync_prefix}sync request {job.name}({job.args}) to `{job.plugin_name}`'
    tb = format_exception(exc).join_lines
    ribo_log.error(f'fatal error in {desc}')
    exc_logger = ribo_log.error if amino.development else ribo_log.debug
    exc_logger(f'{desc} failed:\n{tb}')
    return 1


def resend_send() -> NS[PluginState[D], Tuple[PrioQueue[Message], DispatchResult]]:
    return NS.inspect(lambda state: process_message(state.messages, send_message))


@do(NS[PluginState[D], DispatchResult])
def resend_unpack(result: Tuple[PrioQueue[Message], DispatchResult]) -> Do:
    messages, trs = result
    yield NS.modify(__.copy(messages=messages))
    yield trs


@do(NvimIO[PluginState[D]])
def resend_loop(holder: PluginStateHolder) -> Do:
    @do(NvimIO[PluginState[D]])
    def loop() -> Do:
        messages = holder.messages.join_comma
        yield exclusive_dispatch(holder, resend_send, resend_unpack, f'resend {messages}')
        yield resend_loop(holder)
    yield loop() if holder.has_messages else NvimIO.pure(holder.state)


@do(NvimIO[PluginState[D]])
def execute_async_loop(job: DispatchJob, dispatch: DispatchAsync) -> Do:
    yield execute(job, dispatch, async_runner)
    yield resend_loop(job.state)


@do(NvimIO[Any])
def execute_sync_loop(job: DispatchJob, dispatch: DispatchAsync) -> Do:
    yield execute(job, dispatch, async_runner)
    yield resend_loop(job.state)


def execute_async(job: DispatchJob, dispatch: DispatchAsync) -> NvimIO[None]:
    return fork_job(lambda: execute_async_loop(job, dispatch), job)


@do(NvimIO[Any])
def execute_dispatch_job(job: DispatchJob) -> Do:
    dispatch = yield NvimIO.from_maybe(job.dispatches.lift(job.name), 'no handler')
    relay = execute_sync if dispatch.sync else execute_async
    yield relay(job, dispatch)


__all__ = ('execute_dispatch_job', 'execute_async_loop', 'execute_sync', 'execute')



from typing import Callable, Type, Generic, Any, TypeVar

from amino import List, Lists, Map, _, Boolean
from amino.func import flip
from amino.util.string import ToStr, camelcase
from amino.dat import ADT
from amino.boolean import false

from ribosome.request.rpc import RpcHandlerSpec
from ribosome.logging import Logging
from ribosome.request.handler.dispatcher import RequestDispatcher, MsgDispatcher, TransDispatcher
from ribosome.request.handler.prefix import PrefixStyle, Short
from ribosome.request.handler.method import RpcMethod, CmdMethod, FunctionMethod, AutocmdMethod
from ribosome.trans.message_base import Message

B = TypeVar('B')
D = TypeVar('D')
M = TypeVar('M', bound=Message)
RHS = TypeVar('RHS', bound=RpcHandlerSpec)
DP = TypeVar('DP', bound=RequestDispatcher)
Meth = TypeVar('Meth', bound=RpcMethod)


class RequestHandler(Generic[Meth, DP], ADT['RequestHandler'], Logging):

    def __init__(
            self,
            method: Meth,
            dispatcher: DP,
            name: str,
            prefix: PrefixStyle,
            internal: Boolean,
            sync: Boolean,
            extra_options: Map[str, Any],
    ) -> None:
        self.method = method
        self.dispatcher = dispatcher
        self.name = name
        self.prefix = prefix
        self.internal = internal
        self.sync = sync
        self.extra_options = extra_options

    @staticmethod
    def msg_cmd(msg: Type[M]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(CmdMethod(), MsgDispatcher(msg))

    @staticmethod
    def msg_fun(msg: Type[M]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(FunctionMethod(), MsgDispatcher(msg))

    @staticmethod
    def msg_autocmd(msg: Type[M]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(AutocmdMethod(), MsgDispatcher(msg))

    @staticmethod
    def json_msg_cmd(msg: Type[M]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(CmdMethod(), MsgDispatcher(msg))

    @staticmethod
    def trans_cmd(func: Callable[..., B]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(CmdMethod(), TransDispatcher(func))

    @staticmethod
    def trans_function(func: Callable[..., B]) -> 'RequestHandlerBuilder':
        return RequestHandlerBuilder(FunctionMethod(), TransDispatcher(func))

    @property
    def allow_sync(self) -> Boolean:
        return self.dispatcher.allow_sync

    @property
    def options(self) -> Map[str, Any]:
        return self.extra_options ** Map(sync=self.sync)

    @property
    def method_str(self) -> str:
        return f'{self.method.method}:{self.name}'

    def vim_cmd_name(self, name: str, prefix: str) -> str:
        pre = name if self.prefix.full else prefix if self.prefix.short else ''
        return f'{camelcase(pre)}{camelcase(self.name)}'

    def spec(self, name: str, prefix: str) -> RpcHandlerSpec:
        return RpcHandlerSpec.cons(self.method.spec_type, self.sync, self.vim_cmd_name(name, prefix), self.options,
                                   self.method_str, True)

    @property
    def legacy(self) -> bool:
        return isinstance(self.dispatcher, MsgDispatcher)


class RequestHandlerBuilder(Generic[Meth, DP]):

    def __init__(self, method: Meth, dispatcher: DP) -> None:
        self.method = method
        self.dispatcher = dispatcher

    def __call__(
            self,
            name: str=None,
            prefix: PrefixStyle=Short(),
            internal: Boolean=false,
            sync: Boolean=false,
            **options: Any
    ) -> RequestHandler:
        name1 = name or self.dispatcher.name
        return RequestHandler(self.method, self.dispatcher, name1, prefix, internal, sync, Map(options))


class RequestHandlers(ToStr):

    @staticmethod
    def cons(*handlers: RequestHandler) -> 'RequestHandlers':
        return RequestHandlers(Map(Lists.wrap(handlers).apzip(_.name).map2(flip)))

    def __init__(self, handlers: Map[str, RequestHandler]) -> None:
        self.handlers = handlers

    def _arg_desc(self) -> List[str]:
        return List(str(self.handlers))


__all__ = ('RequestHandler', 'RequestHandlers')



from kallikrein import Expectation

from ribosome.plugin import PluginSettings
from ribosome.config import SimpleData
from ribosome.test.integration.klk import AutoPluginIntegrationKlkSpec

from integration._support.autocmd import val


class AutocmdSpec(AutoPluginIntegrationKlkSpec[PluginSettings, SimpleData]):
    '''
    execute handler when triggering an autocmd $autocmd
    '''

    @property
    def plugin_prefix(self) -> str:
        return 'plug'

    def module(self) -> str:
        return 'integration._support.autocmd'

    def _pre_start(self) -> None:
        super()._pre_start()
        self.vim.vars.set_p('components', ['core'])

    def autocmd(self) -> Expectation:
        self.vim.cmd_once_defined('PlugStage1')
        self.cmd_sync('Msg1')
        self.vim.doautocmd('VimResized')
        return self.var_becomes('msg_cmd_success', val) & self.var_becomes('autocmd_success', val)

__all__ = ('AutocmdSpec',)



from ribosome.nvim.components import HasNvim, NvimFacade, AsyncVimProxy, Tab, Buffer, Window
from ribosome.nvim.io import NvimIO

__all__ = ('NvimFacade', 'HasNvim', 'NvimIO', 'ScratchBuilder',
           'AsyncVimProxy', 'Tab', 'ScratchBuffer', 'Buffer', 'Window')



import functools
from typing import TypeVar, Callable

import neovim

from ribosome.request.legacy import LegacyRequestHandler
from ribosome.trans.message_base import Message
from ribosome.request.handler.method import RpcMethod, AutocmdMethod

M = TypeVar('M', bound=Message)


class Autocmd(LegacyRequestHandler):

    @property
    def desc(self) -> str:
        return 'autocmd'

    @property
    def neovim_autocmd(self) -> Callable:
        @neovim.autocmd(self.vim_name, **self.kw)
        @functools.wraps(self._fun)
        def neovim_autocmd_wrapper(obj, *rpc_args):
            return self.dispatch(obj, rpc_args)
        return neovim_autocmd_wrapper

    @property
    def method(self) -> RpcMethod:
        return AutocmdMethod()


def autocmd(**kw):
    def neovim_autocmd_decorator(fun):
        handler = Autocmd(fun, **kw)
        fun = handler.neovim_autocmd
        fun.spc = handler.spec
        return fun
    return neovim_autocmd_decorator

__all__ = ('autocmd',)



import sys
import logging


def remove_path(p: str) -> None:
    if p in sys.path:
        sys.path.remove(p)


def stage1(log: logging.Logger) -> int:
    try:
        remove_path('')
        remove_path('.')
        from amino import Lists
        from ribosome.host import start_file
        log.debug(f'ribosome_start_plugin: {sys.argv}, {sys.path}')
        def no_args() -> int:
            log.error(f'ribosome_start_plugin: missing argument for plugin file')
            return 1
        return Lists.wrap(sys.argv).lift(1).cata(start_file, no_args)
    except Exception as e:
        log.caught_exception_error(f'starting plugin with {sys.argv}', e)
        return 1


def start_plugin() -> int:
    from amino import with_log
    return with_log(stage1)


__all__ = ('start_plugin',)



in_vim = True

from ribosome.nvim import NvimFacade
from ribosome.plugin import NvimStatePlugin, NvimPlugin, AutoPlugin
from ribosome.request.command import command
from ribosome.request.function import function

__all__ = ('command', 'NvimFacade', 'NvimStatePlugin', 'NvimPlugin', 'in_vim', 'function', 'AutoPlugin')



import abc
from typing import Callable, Type, Generic, Any, TypeVar

from amino import List, Nil, Boolean
from amino.util.string import snake_case
from amino.algebra import Algebra

from ribosome.logging import Logging
from ribosome.trans.handler import FreeTransHandler
from ribosome.trans.message_base import Message

B = TypeVar('B')
D = TypeVar('D')
M = TypeVar('M', bound=Message)


class RequestDispatcher(Algebra, Logging, base=True):

    @abc.abstractproperty
    def args(self) -> List[Any]:
        ...

    @abc.abstractproperty
    def name(self) -> str:
        ...

    @property
    def allow_sync(self) -> Boolean:
        return Boolean.isinstance(self, SyncRequestDispatcher)


class SyncRequestDispatcher(RequestDispatcher):
    pass


class AsyncRequestDispatcher(RequestDispatcher):
    pass


class MsgDispatcher(Generic[M], AsyncRequestDispatcher):

    def __init__(self, msg: Type[M]) -> None:
        self.msg = msg

    def _arg_desc(self) -> List[str]:
        return List(self.msg.__name__)

    @property
    def args(self) -> List[Any]:
        return List(self.msg)

    @property
    def name(self) -> str:
        return snake_case(self.msg.__name__)


class TransDispatcher(Generic[B], SyncRequestDispatcher):

    def __init__(self, handler: FreeTransHandler[D, B]) -> None:
        self.handler = handler

    @property
    def args(self) -> List[Any]:
        return Nil

    def _arg_desc(self) -> List[str]:
        return List(self.name)

    @property
    def name(self) -> str:
        return self.handler.name


class FunctionDispatcher(SyncRequestDispatcher):

    def __init__(self, fun: Callable) -> None:
        self.fun = fun

    def _arg_desc(self) -> List[str]:
        return List()

    @property
    def args(self) -> List[Any]:
        return List()

    @property
    def name(self) -> str:
        return snake_case(self.fun.__name__)


__all__ = ('RequestDispatcher', 'SyncRequestDispatcher', 'AsyncRequestDispatcher', 'MsgDispatcher', 'TransDispatcher',
           'FunctionDispatcher')



from typing import TypeVar, Any, Generic, Type

from amino import Map, List, Boolean, Nil, Either, Left, Right, __, _
from amino.dat import Dat
from amino.algebra import Algebra
from amino.do import do, Do

from ribosome.nvim import NvimIO
from ribosome.logging import Logging
from ribosome.request.args import ArgValidator, ParamsSpec
from ribosome.plugin_state import PluginState, PluginStateHolder
from ribosome.dispatch.data import Legacy, DispatchReturn, Internal, Trans, SendMessage, DispatchResult
from ribosome.nvim.io import NvimIOState
from ribosome.trans.message_base import Message
from ribosome.dispatch.transform import TransValidator, validate_trans_action
from ribosome.trans.send_message import send_message, transform_data_state
from ribosome.trans.handler import FreeTransHandler

NP = TypeVar('NP')
D = TypeVar('D')
DP = TypeVar('DP', bound=Algebra)
Res = NvimIOState[PluginState[D], DispatchResult]


def execute_data_trans(handler: FreeTransHandler) -> Res:
    return transform_data_state(validate_trans_action(handler.run()))


def run_trans(trans: Trans, args: List[Any]) -> Res:
    handler = trans.handler.dispatcher.handler(*args)
    return execute_data_trans(handler)


def run_internal(trans: Trans, args: List[Any]) -> Res:
    handler = trans.handler.dispatcher.handler(args)
    result = handler.run()
    validator = TransValidator(trans.name)
    return validator.validate(result)


def cons_message(tpe: Type[Message], args: List[Any], cmd_name: str, method: str) -> Either[str, Message]:
    validator = ArgValidator(ParamsSpec.from_function(tpe.__init__))
    return Right(tpe(*args)) if validator.validate(args) else Left(validator.error(args, method, cmd_name))


class RunDispatch(Generic[D, NP], Logging):

    def internal(self, dispatch: Internal) -> Res:
        return run_internal(dispatch, self.args)

    def trans(self, dispatch: Trans) -> Res:
        return run_trans(dispatch, self.args)


class RunDispatchSync(RunDispatch):

    def __init__(self, args: List[Any]) -> None:
        self.args = args

    @do(NvimIOState[PluginState[D], DispatchResult])
    def legacy(self, dispatch: Legacy) -> Do:
        plugin = yield NvimIOState.inspect(_.plugin)
        result = dispatch.handler.func(plugin, *self.args)
        yield NvimIOState.pure((DispatchResult(DispatchReturn(result), Nil)))


class RunDispatchAsync(RunDispatch):

    def __init__(self, args: List[Any]) -> None:
        self.args = args

    @do(NvimIOState[PluginState[D], DispatchResult])
    def send_message(self, dispatch: SendMessage) -> Do:
        cmd_name = yield NvimIOState.inspect(__.config.vim_cmd_name(dispatch.handler))
        msg_e = cons_message(dispatch.msg, self.args, cmd_name, dispatch.method)
        yield msg_e.cata(DispatchResult.error_nio, send_message)


def invalid_dispatch(data: Any) -> NvimIO[Any]:
    return NvimIO.failed(f'invalid type passed to `RunDispatch`: {data}')


class DispatchJob(Generic[DP], Dat['DispatchJob']):

    def __init__(
            self,
            dispatches: Map[str, DP],
            state: PluginStateHolder[D],
            name: str,
            args: List[Any],
            sync: bool,
            plugin_name: str,
            bang: Boolean,
    ) -> None:
        self.dispatches = dispatches
        self.state = state
        self.name = name
        self.args = args
        self.sync = sync
        self.plugin_name = plugin_name
        self.bang = Boolean(bang)


__all__ = ('PluginState', 'PluginStateHolder', 'invalid_dispatch', 'DispatchJob')



from functools import wraps
from typing import Callable, Any

from amino import env


def unite(f: Callable[[Any, str], None]) -> Callable[[Any], None]:
    @wraps(f)
    def wrapper(self: Any) -> None:
        def go(unite: str) -> Any:
            self.vim.options.amend_l('rtp', [unite])
            self.vim.cmd('source {}/plugin/*.vim'.format(unite))
            self.vim.cmd('source {}/plugin/unite/*.vim'.format(unite))
            self.vim.cmd('source {}/syntax/*.vim'.format(unite))
            return f(self)
        return env['UNITE_DIR'] / go | None
    return wrapper

__all__ = ('unite',)




class LegacyData:
    # vim_facade = maybe_field((NvimFacade, AsyncVimProxy))
    # sub_states = field(Map, initial=Map())
    # debug = bool_field()

    # def sub_state_m(self, name):
    #     return self.sub_states.get(name)

    # def sub_state(self, name, default: Callable[[], Any]):
    #     return self.sub_state_m(name) | default

    # def with_sub_state(self, name, state):
    #     new_states = self.sub_states + (name, state)
    #     return self.set(sub_states=new_states)

__all__ = ('LegacyData',)



import neovim

from ribosome.request.legacy import LegacyRequestHandler
from ribosome.request.handler.method import RpcMethod, FunctionMethod

from amino import Map


class Function(LegacyRequestHandler):

    @property
    def desc(self):
        return 'function'

    @property
    def default_kw(self):
        return Map(sync=True)

    @property
    def neovim_fun(self):
        @neovim.function(self.vim_name, **self.kw)
        def neovim_fun_wrapper(obj, *rpc_args):
            return self.dispatch(obj, rpc_args)
        return neovim_fun_wrapper

    @property
    def method(self) -> RpcMethod:
        return FunctionMethod()


def function(**kw):
    def neovim_fun_decorator(fun):
        handler = Function(fun, **kw)
        fun.spec = handler.spec
        nfun = handler.neovim_fun
        nfun.spec = handler.spec
        return nfun
    return neovim_fun_decorator

__all__ = ('function',)



from amino import Map, List
from ribosome.config import Config, RequestHandler
from ribosome.trans.message_base import Message
from ribosome.trans.messages import Stage1
from ribosome.trans.api import trans
from ribosome.dispatch.component import Component


class Pub(Message['Pub']):
    pass


class P1(Message['P1']):
    pass


class P2(Message['P2']):
    pass


class Target(Message['Target']):
    pass


class T1(Message['T1']):
    pass


class T2(Message['T2']):
    pass


class T3(Message['T3']):
    pass


class Seq(Message['Seq']):
    pass


class S1(Message['S1']):
    pass


class S2(Message['S2']):
    pass


class S3(Message['S3']):
    pass


class P(Component):

    @trans.msg.unit(Stage1)
    def stage_1(self) -> None:
        pass

    @trans.msg.one(Target)
    def target(self) -> Message:
        return T1().to('q')

    @trans.msg.one(T1)
    def t1(self) -> Message:
        return T2()

    @trans.msg.multi(Seq)
    def seq(self) -> List[Message]:
        return List(S1().at(1), S2().at(0.1))


class Q(Component):

    @trans.msg.one(P1)
    def p1(self) -> Message:
        return P2()

    @trans.msg.one(T1)
    def t1(self) -> Message:
        return T3()

    @trans.msg.one(S2)
    def s1(self) -> Message:
        return S3().at(0.6)


@trans.msg.one(Pub)
def pub(msg: Pub) -> Message:
    return P1()


config = Config.cons(
    name='envl',
    components=Map(p=P, q=Q),
    request_handlers=List(
        RequestHandler.trans_cmd(pub)('pub'),
        RequestHandler.msg_cmd(Target)('target'),
        RequestHandler.msg_cmd(Seq)('Seq'),
    ),
    core_components=List('p', 'q')
)

__all__ = ('config',)



from typing import TypeVar, Type, Callable, Any
from types import ModuleType

from neovim.msgpack_rpc import MsgpackStream, AsyncSession, Session
from neovim.api import Nvim
from neovim.msgpack_rpc.event_loop.base import BaseEventLoop
from neovim.msgpack_rpc.event_loop.uv import UvEventLoop

from amino import Either, _, L, Maybe, amino_log, Logger, __, Path, Map, List, Boolean, Nil, Just
from amino.either import ImportFailure
from amino.func import Val
from amino.logging import amino_root_file_logging
from amino.do import do, Do
from amino.dat import Dat
from amino.algebra import Algebra
from amino.state import EitherState, MaybeState
from amino.json import dump_json
from amino.boolean import false
from amino.util.string import decode
from amino.mod import instance_from_module, cls_from_module

from ribosome import NvimPlugin
from ribosome import options
from ribosome.components.scratch import Mapping
from ribosome.config import Config
from ribosome.dispatch.data import Legacy, SendMessage, Trans, Internal, DispatchSync, DispatchAsync, Dispatch, DIO
from ribosome.dispatch.execute import execute_dispatch_job, request_error, execute_io
from ribosome.dispatch.resolve import ComponentResolver
from ribosome.dispatch.run import DispatchJob
from ribosome.logging import ribo_log, nvim_logging
from ribosome.nvim import NvimFacade, NvimIO
from ribosome.plugin import plugin_class_from_config
from ribosome.plugin_state import PluginState, PluginStateHolder, DispatchConfig
from ribosome.request.handler.dispatcher import MsgDispatcher
from ribosome.request.handler.handler import RequestHandler
from ribosome.request.handler.prefix import Full
from ribosome.request.rpc import rpc_handler_functions, define_handlers, RpcHandlerSpec
from ribosome.trans.api import trans
from ribosome.trans.messages import ShowLogInfo, UpdateState, Quit, Stage1
from ribosome.trans.queue import PrioQueue
from ribosome.nvim.io import NS
from ribosome.trans.handler import TransComplete

Loop = TypeVar('Loop', bound=BaseEventLoop)
NP = TypeVar('NP', bound=NvimPlugin)
D = TypeVar('D')
B = TypeVar('B')
C = TypeVar('C', bound=Config)
R = TypeVar('R')
DP = TypeVar('DP', bound=Dispatch)
RDP = TypeVar('RDP', bound=Algebra)


class HostConfig(Dat['HostConfig']):

    def __init__(
            self,
            sync_dispatch: Map[str, DispatchSync],
            async_dispatch: Map[str, DispatchAsync],
            config: DispatchConfig,
            plugin_class: Maybe[Type[NP]],
    ) -> None:
        self.sync_dispatch = sync_dispatch
        self.async_dispatch = async_dispatch
        self.config = config
        self.plugin_class = plugin_class

    @property
    def name(self) -> str:
        return self.config.name

    @property
    def dispatch(self) -> List[Dispatch]:
        return self.sync_dispatch.v + self.async_dispatch.v

    @property
    def specs(self) -> List[RpcHandlerSpec]:
        return self.dispatch / __.spec(self.config.name, self.config.prefix)


def dispatch_job(sync: bool,
                 dispatches: Map[str, RDP],
                 state: PluginStateHolder[D],
                 name: str,
                 prefix: str,
                 args: tuple) -> DispatchJob:
    decoded_args = decode(args)
    fun_args = decoded_args.head | Nil
    bang = decoded_args.lift(1).contains(1)
    return DispatchJob(dispatches, state, decode(name), fun_args, sync, prefix, bang)


def request_handler(vim: NvimFacade,
                    sync: bool,
                    dispatches: Map[str, RDP],
                    state: PluginStateHolder[D],
                    config: Config) -> Callable[[str, tuple], Any]:
    sync_prefix = '' if sync else 'a'
    def handle(name: str, args: tuple) -> Do:
        job = dispatch_job(sync, dispatches, state, name, config.prefix, args)
        amino_log.debug(f'dispatching {sync_prefix}sync request: {job.name}({job.args})')
        result = (
            execute_dispatch_job(job)
            .attempt(vim)
            .value_or(L(request_error)(job, _))
        )
        if sync:
            ribo_log.debug(f'request `{job.name}` completed: {result}')
        return vim.encode_vim_data(result)
    return handle


@do(NvimIO[PluginState[D]])
def init_state(host_config: HostConfig) -> Do:
    dispatch_config = host_config.config
    data = dispatch_config.config.state()
    components = yield ComponentResolver(dispatch_config.config).run
    plugin = yield NvimIO.delay(host_config.plugin_class)
    log_handler = yield NvimIO.delay(nvim_logging)
    yield NvimIO.pure(PluginState.cons(dispatch_config, data, plugin, components, PrioQueue.empty, Nil, Just(log_handler)))


# TODO can vim be injected into each request handling process?
@do(NvimIO[int])
def run_session(session: Session, host_config: HostConfig) -> Do:
    yield define_handlers(host_config.specs, host_config.name, host_config.name)
    state = yield init_state(host_config)
    holder = PluginStateHolder.cons(state)
    yield NvimIO.delay(
        lambda vim:
        session.run(
            request_handler(vim, True, host_config.sync_dispatch, holder, host_config.config),
            request_handler(vim, False, host_config.async_dispatch, holder, host_config.config)
        )
    )
    yield NvimIO.pure(0)


def run_loop(session: Session, prefix: str, host_config: HostConfig) -> int:
    vim = NvimFacade(Nvim.from_session(session), host_config.name)
    return run_session(session, host_config).attempt(vim).get_or_raise()


def session(*args: str, loop: Loop=UvEventLoop, transport_type: str='stdio', **kwargs: str) -> Session:
    return Session(AsyncSession(MsgpackStream(loop(transport_type, *args, **kwargs))))


def start_host(prefix: str, host_config: HostConfig) -> int:
    return run_loop(session(), prefix, host_config)


def plugin_cls_from_module(mod: ModuleType) -> Either[str, Type[NP]]:
    return cls_from_module(mod, NvimPlugin)


def config_from_module(mod: ModuleType) -> Either[str, Type[C]]:
    return instance_from_module(mod, Config)


log_initialized = False


def nvim_log() -> Logger:
    global log_initialized
    if not log_initialized:
        NvimFacade.stdio_with_logging('ribosome_start_host')
        log_initialized = True
    return ribo_log


def plugin_class_dispatchers(cls: Type[NP]) -> List[DispatchSync]:
    return rpc_handler_functions(cls) / Legacy


def config_dispatchers(config: Config) -> List[DispatchAsync]:
    def choose(name: str, handler: RequestHandler) -> DispatchAsync:
        tpe = (
            SendMessage
            if isinstance(handler.dispatcher, MsgDispatcher) else
            Internal
            if handler.internal else
            Trans
        )
        return tpe(handler)
    return config.request_handlers.handlers.map2(choose)


@trans.free.result(trans.st)
@do(EitherState[PluginState[D], str])
def message_log() -> Do:
    yield EitherState.inspect_f(__.message_log.traverse(dump_json, Either))


@trans.free.unit(trans.st)
def set_log_level(self, level: str) -> None:
    handler = yield MaybeState.inspect_f(_.file_log_handler)
    handler.setLevel(level)


# TODO change remaining handlers to trans
message_log_handler = RequestHandler.trans_function(message_log)('message_log', Full(), internal=True, sync=True)
set_log_level_handler = RequestHandler.trans_function(set_log_level)(prefix=Full(), internal=True)
show_log_info_handler = RequestHandler.msg_cmd(ShowLogInfo)('show_log_info', prefix=Full(), internal=True)
update_state_handler = RequestHandler.json_msg_cmd(UpdateState)('update_state', prefix=Full())
mapping_handler = RequestHandler.msg_fun(Mapping)('mapping', prefix=Full())
stage_1_handler = RequestHandler.msg_cmd(Stage1)('stage_1', prefix=Full())
quit_handler = RequestHandler.msg_cmd(Quit)(prefix=Full())


def internal_dispatchers(config: Config) -> List[Dispatch]:
    return List(
        Internal(message_log_handler),
        Internal(set_log_level_handler),
        SendMessage(show_log_info_handler),
        SendMessage(update_state_handler),
        SendMessage(mapping_handler),
        SendMessage(stage_1_handler),
        SendMessage(quit_handler),
    )


def host_config_1(
        config: Config,
        cls: Type[NP],
        debug: Boolean,
        io_executor: Callable[[DIO], NS[PluginState[D], TransComplete]]=None,
) -> HostConfig:
    name = config.name
    cls_dispatchers = plugin_class_dispatchers(cls)
    cfg_dispatchers = config_dispatchers(config)
    int_dispatchers = internal_dispatchers(config)
    with_method = lambda ds: Map(ds.map(lambda d: (d.spec(config.name, config.prefix).rpc_method(name), d)))
    dispatches = int_dispatchers + cfg_dispatchers + cls_dispatchers
    sync_dispatch, async_dispatch = dispatches.filter(_.sync), dispatches.filter(_.async)
    dispatch_config = DispatchConfig.cons(config, io_executor)
    return HostConfig(with_method(sync_dispatch), with_method(async_dispatch), dispatch_config, cls)


class SyntheticPlugin:
    pass


def host_config(
        config: Config,
        plugin_mixin: Either[str, Type[NP]],
        debug: Boolean=false,
        io_executor: Callable[[DIO], NS[PluginState[D], TransComplete]]=None,
) -> HostConfig:
    sup = plugin_mixin | Val(SyntheticPlugin)
    cls = plugin_class_from_config(config, sup, debug)
    return host_config_1(config, cls, debug, io_executor)


def start_config_stage_2(sup: Either[str, Type[NP]], config: Config) -> int:
    debug = options.development.exists
    amino_log.debug(f'starting plugin from {config}, debug: {debug}')
    return start_host(config.name, host_config(config, sup, debug))


def error(msg: str) -> int:
    amino_log.error(msg)
    nvim_log().error(msg)
    return 1


def import_error(e: ImportFailure, desc: str) -> int:
    return error(e.expand.join_lines)


def exception(e: Exception, desc: str) -> int:
    f = __.caught_exception_error(f'starting host from {desc}', e)
    f(amino_log)
    f(nvim_log())
    return 1


def start_config_stage_1(mod: ModuleType) -> int:
    nvim_log()
    plugin_cls = plugin_cls_from_module(mod)
    return config_from_module(mod).cata(error, L(start_config_stage_2)(plugin_cls, _))


def setup_log() -> None:
    amino_root_file_logging()


def start_from(source: str, importer: Callable[[str], Either[ImportFailure, ModuleType]], desc: str) -> int:
    try:
        setup_log()
        amino_log.debug(f'start_{desc}: {source}')
        return importer(source).cata(L(import_error)(_, source), start_config_stage_1)
    except Exception as e:
        return exception(e, source)


def start_module(mod: str) -> int:
    return start_from(mod, Either.import_module, 'module')


def start_file(path: str) -> int:
    p = Path(path)
    file = p / '__init__.py' if p.is_dir() else p
    return start_from(str(file), Either.import_file, 'file')


__all__ = ('start_host', 'start_module', 'start_file')



import abc
from typing import TypeVar, Generic, Union

from amino import Maybe, List, __, Nil, Either, L, _
from amino.state import StateT, MaybeState, EitherState, IdState
from amino.id import Id
from amino.dispatch import dispatch_alg
from amino.tc.base import TypeClass, F
from amino.util.string import blue

from ribosome.logging import Logging
from ribosome.nvim.io import NvimIOState, NS
from ribosome.nvim import NvimIO
from ribosome.dispatch.data import (DispatchResult, DispatchUnit, DispatchError, DispatchReturn, DispatchIO, DIO,
                                    DispatchDo)
from ribosome.trans.action import (Transit, Propagate, TransUnit, TransResult, TransFailure, TransAction, TransIO,
                                   TransDo)
from ribosome.trans.message_base import Message
from ribosome.trans.handler import TransComplete


D = TypeVar('D')
R = TypeVar('R')
G = TypeVar('G', bound=F)
I = TypeVar('I')


class TransformTransState(Generic[G], TypeClass):

    @abc.abstractmethod
    def transform(self, st: StateT[G, D, DispatchResult]) -> NvimIOState[D, DispatchResult]:
        ...

    def run(self, st: StateT[G, D, List[Message]]) -> NvimIOState[D, DispatchResult]:
        return self.transform(st)


class TransformIdState(TransformTransState[Id], tpe=IdState):

    def transform(self, st: IdState[D, DispatchResult]) -> NvimIOState[D, DispatchResult]:
        return NvimIOState.from_id(st)


class TransformMaybeState(TransformTransState[Maybe], tpe=MaybeState):

    def transform(self, st: MaybeState[D, DispatchResult]) -> NvimIOState[D, DispatchResult]:
        return NvimIOState.apply(lambda s: NvimIO.pure(st.run(s) | (s, DispatchResult(DispatchUnit(), Nil))))


class TransformEitherState(TransformTransState[Either], tpe=EitherState):

    def transform(self, st: EitherState[D, DispatchResult]) -> NvimIOState[D, DispatchResult]:
        return NvimIOState.apply(
            lambda s: NvimIO.pure(st.run(s).value_or(lambda err: (s, DispatchResult(DispatchError.cons(err), Nil))))
        )


class TransformNvimIOState(TransformTransState[NvimIO], tpe=NvimIOState):

    def transform(self, st: NvimIOState[D, DispatchResult]) -> NvimIOState[D, DispatchResult]:
        return st


class TransValidator(Logging):

    def __init__(self, name: str) -> None:
        self.name = name

    def failure(self, error: Union[str, Exception]) -> NvimIOState[D, DispatchResult]:
        return NvimIOState.pure(DispatchResult(DispatchError.cons(f'trans {blue(self.name)}: {error}'), Nil))

    def validate_transit(self, action: Transit) -> NvimIOState[D, DispatchResult]:
        def loop(tts: TransformTransState) -> NS[D, DispatchResult]:
            return tts.run(action.trans).map(L(TransComplete)(self.name, _)).flat_map(validate_trans_action)
        return (
            TransformTransState.e_for(action.trans)
            .map(loop)
            .value_or(self.failure)
        )

    def validate_propagate(self, action: Propagate) -> NvimIOState[D, DispatchResult]:
        return NvimIOState.pure(DispatchResult(DispatchUnit(), action.messages))

    def validate_trans_unit(self, action: TransUnit) -> NvimIOState[D, DispatchResult]:
        return NvimIOState.pure(DispatchResult(DispatchUnit(), action.messages))

    def validate_trans_result(self, action: TransResult) -> NvimIOState[D, DispatchResult]:
        return NvimIOState.pure(DispatchResult(DispatchReturn(action.data), Nil))

    def validate_trans_failure(self, action: TransFailure) -> NvimIOState[D, DispatchResult]:
        return self.failure(action.message)

    def validate_trans_io(self, action: TransIO[I]) -> NvimIOState[D, DispatchResult]:
        output = DIO.cons(action.io).cata(DispatchError.cons, DispatchIO)
        return NvimIOState.pure(DispatchResult(output, Nil))

    def validate_trans_do(self, action: TransDo) -> NvimIOState[D, DispatchResult]:
        return NvimIOState.pure(DispatchResult(DispatchDo(action), Nil))


def validate_trans_action(tc: TransComplete) -> NS[D, DispatchResult]:
    val = dispatch_alg(TransValidator(tc.name), TransAction, 'validate_')
    return val(tc.action)

__all__ = ('validate_trans_action',)



import abc
import inspect
import traceback
import typing
from typing import TypeVar, Callable, Any, Generic, Generator, Union, Tuple
from threading import Thread

from fn.recur import tco

from amino.tc.base import ImplicitInstances, F, TypeClass, tc_prop
from amino.lazy import lazy
from amino.tc.monad import Monad
from amino import Either, __, IO, Maybe, Left, Eval, L, List, Nothing, Right, Lists, _, Just, options
from amino.state import tcs, StateT, State
from amino.func import CallByName
from amino.do import do
from amino.io import safe_fmt
from amino.util.string import ToStr
from amino.util.fun import lambda_str
from amino.util.exception import sanitize_tb, format_exception
from amino.dat import ADT

from ribosome.nvim.components import NvimFacade

A = TypeVar('A')
B = TypeVar('B')
S = TypeVar('S')


class NvimIOException(Exception):
    remove_pkgs = List('amino', 'fn')

    def __init__(self, f, stack, cause) -> None:
        self.f = f
        self.stack = List.wrap(stack)
        self.cause = cause

    @property
    def location(self):
        files = List('io', 'anon', 'instances/io', 'tc/base', 'nvim/io')
        def filt(entry, name):
            return entry.filename.endswith('/{}.py'.format(name))
        stack = self.stack.filter_not(lambda a: files.exists(L(filt)(a, _)))
        pred = (lambda a: not NvimIOException.remove_pkgs
                .exists(lambda b: '/{}/'.format(b) in a.filename))
        return stack.find(pred)

    @property
    def format_stack(self) -> List[str]:
        rev = self.stack.reversed
        def remove_recursion(i):
            pre = rev[:i + 1]
            post = rev[i:].drop_while(__.filename.endswith('/amino/io.py'))
            return pre + post
        def remove_internal():
            start = rev.index_where(_.function == 'unsafe_perform_sync')
            return start / remove_recursion | rev
        frames = (self.location.to_list if IO.stack_only_location else remove_internal())
        data = frames / (lambda a: a[1:-2] + tuple(a[-2]))
        return sanitize_tb(Lists.wrap(traceback.format_list(list(data))))

    @property
    def lines(self) -> List[str]:
        cause = format_exception(self.cause)
        suf1 = '' if self.stack.empty else ' at:'
        tb1 = (List() if self.stack.empty else self.format_stack)
        return tb1.cons(f'IO exception{suf1}').cat('Cause:') + cause + List(
            '',
            'Callback:',
            f'  {self.f}'
        )

    def __str__(self):
        return self.lines.join_lines


class NvimIOInstances(ImplicitInstances):

    @lazy
    def _instances(self) -> 'amino.map.Map':
        from amino.map import Map
        return Map({Monad: NvimIOMonad()})


class NResult(Generic[A], ADT['NvimIOResult[A]']):
    pass


class NSuccess(Generic[A], NResult[A]):

    def __init__(self, value: A) -> None:
        self.value = value


class NError(Generic[A], NResult[A]):

    def __init__(self, error: str) -> None:
        self.error = error


class NFatal(Generic[A], NResult[A]):

    def __init__(self, exception: Exception) -> None:
        self.exception = exception


# TODO make this an ADT, create NvimIOFailed that circumvents exception throwing
class NvimIO(Generic[A], F[A], ToStr, implicits=True, imp_mod='ribosome.nvim.io', imp_cls='NvimIOInstances'):
    debug = options.io_debug.exists

    @staticmethod
    def wrap_either(f: Callable[[NvimFacade], Either[B, A]]) -> 'NvimIO[A]':
        return NvimIO.delay(lambda a: f(a).get_or_raise())

    @staticmethod
    def from_either(e: Either[str, A]) -> 'NvimIO[A]':
        return NvimIO.wrap_either(lambda v: e)

    @staticmethod
    def from_maybe(e: Maybe[A], error: CallByName) -> 'NvimIO[A]':
        return NvimIO.from_either(e.to_either(error))

    @staticmethod
    def cmd_sync(cmdline: str, verbose=False) -> 'NvimIO[str]':
        return NvimIO.wrap_either(__.cmd_sync(cmdline, verbose=verbose))

    @staticmethod
    def cmd(cmdline: str, verbose=False) -> 'NvimIO[str]':
        return NvimIO.wrap_either(__.cmd(cmdline, verbose=verbose))

    @staticmethod
    def call(name: str, *args: Any, **kw: Any) -> 'NvimIO[A]':
        return NvimIO.wrap_either(__.call(name, *args, **kw))

    @staticmethod
    def call_once_defined(name: str, *args: Any, **kw: Any) -> 'NvimIO[A]':
        return NvimIO.wrap_either(__.call_once_defined(name, *args, **kw))

    @staticmethod
    def exception(exc: Exception) -> 'NvimIO[A]':
        return NvimIO.from_either(Left(exc))

    @staticmethod
    def failed(msg: str) -> 'NvimIO[A]':
        return NvimIO.exception(Exception(msg))

    @staticmethod
    def error(msg: str) -> 'NvimIO[A]':
        return NvimIOError(msg)

    @staticmethod
    def from_io(io: IO[A]) -> 'NvimIO[A]':
        return NvimIO.delay(lambda a: io.attempt.get_or_raise())

    @staticmethod
    def fork(f: Callable[[NvimFacade], None]) -> 'NvimIO[None]':
        return NvimIO.delay(lambda v: Thread(target=f, args=(v,)).start())

    @staticmethod
    def delay(f: Callable[..., A], *a: Any, **kw: Any) -> 'NvimIO[A]':
        def g(vim: NvimFacade) -> A:
            return Pure(f(vim, *a, **kw))
        return Suspend(g, safe_fmt(f, ('vim',) + a, kw))

    @staticmethod
    def suspend(f: Callable[..., 'NvimIO[A]'], *a: Any, **kw: Any) -> 'NvimIO[A]':
        def g(vim: NvimFacade) -> NvimIO[A]:
            return f(vim, *a, **kw)
        return Suspend(g, safe_fmt(f, a, kw))

    @staticmethod
    def pure(a: A) -> 'NvimIO[A]':
        return Pure(a)

    @abc.abstractmethod
    def lambda_str(self) -> Eval[str]:
        ...

    @abc.abstractmethod
    def _flat_map(self, f: Callable[[A], 'NvimIO[B]'], ts: Eval[str], fs: Eval[str]) -> 'NvimIO[B]':
        ...

    @abc.abstractmethod
    def step1(self, vim: NvimFacade) -> 'NvimIO[A]':
        ...

    def __init__(self) -> None:
        self.stack = inspect.stack() if NvimIO.debug else []

    def flat_map(self, f: Callable[[A], 'NvimIO[B]'], ts: Maybe[Eval[str]]=Nothing, fs: Maybe[Eval[str]]=Nothing
                 ) -> 'NvimIO[B]':
        ts1 = ts | self.lambda_str
        fs1 = fs | Eval.later(lambda: f'flat_map({lambda_str(f)})')
        return self._flat_map(f, ts1, fs1)

    def _arg_desc(self) -> List[str]:
        return List(self.lambda_str().evaluate())

    def step(self, vim: NvimFacade) -> Union[A, 'NvimIO[A]']:
        try:
            return self.step1(vim)
        except NvimIOException as e:
            raise e
        except Exception as e:
            raise NvimIOException(self.lambda_str().evaluate(), self.stack, e)

    def run(self, vim: NvimFacade) -> A:
        @tco
        def run(t: Union[A, 'NvimIO[A]']) -> Union[Tuple[bool, A], Tuple[bool, Tuple[Union[A, 'NvimIO[A]']]]]:
            if isinstance(t, Pure):
                return True, (t.value,)
            elif isinstance(t, NvimIO):
                return True, (t.step(vim),)
            else:
                return False, t
        return run(self)

    def compute_result(self, vim: NvimFacade) -> NResult[A]:
        try:
            return NSuccess(self.run(vim))
        except NvimIOException as e:
            return NFatal(e)

    def result(self, vim: NvimFacade) -> NResult[A]:
        return (
            NError(self.error)
            if isinstance(self, NvimIOError) else
            self.compute_result(vim)
        )

    def either(self, vim: NvimFacade) -> Either[NvimIOException, A]:
        try:
            return Right(self.run(vim))
        except NvimIOException as e:
            return Left(e)

    def attempt(self, vim: NvimFacade) -> Either[NvimIOException, A]:
        return self.either(vim)

    def with_stack(self, s: typing.List[inspect.FrameInfo]) -> 'IO[A]':
        self.stack = s
        return self

    def unsafe(self, vim: NvimFacade) -> A:
        return self.attempt(vim).get_or_raise()

    def recover(self, f: Callable[[Exception], B]) -> 'NvimIO[B]':
        return NvimIO.delay(self.attempt).map(__.value_or(f))

    @do('NvimIO[A]')
    def ensure(self, f: Callable[[Either[Exception, A]], 'NvimIO[None]']) -> Generator:
        result = yield NvimIO.delay(self.attempt)
        yield f(result)
        yield NvimIO.from_either(result)

    def effect(self, f: Callable[[A], Any]) -> 'NvimIO[A]':
        def wrap(v: NvimFacade) -> A:
            ret = self.run(v)
            f(ret)
            return ret
        return NvimIO.delay(wrap)

    __mod__ = effect

    def error_effect(self, f: Callable[[Exception], None]) -> 'NvimIO[A]':
        return self.ensure(lambda a: NvimIO.delay(lambda v: a.leffect(f)))


class Suspend(Generic[A], NvimIO[A]):

    def __init__(self, thunk: Callable[[NvimFacade], NvimIO[A]], string: Eval[str]) -> None:
        super().__init__()
        self.thunk = thunk
        self.string = string

    def lambda_str(self) -> Eval[str]:
        return self.string

    def step1(self, vim: NvimFacade) -> NvimIO[A]:
        return self.thunk(vim).with_stack(self.stack)

    def _flat_map(self, f: Callable[[A], NvimIO[B]], ts: Eval[str], fs: Eval[str]) -> NvimIO[B]:
        return BindSuspend(self.thunk, f, ts, fs)


class BindSuspend(Generic[A], NvimIO[A]):

    def __init__(self, thunk: Callable[[NvimFacade], NvimIO[A]], f: Callable, ts: Eval[str], fs: Eval[str]) -> None:
        super().__init__()
        self.thunk = thunk
        self.f = f
        self.ts = ts
        self.fs = fs

    def lambda_str(self) -> Eval[str]:
        return (self.ts & self.fs).map2('{}.{}'.format)

    def step1(self, vim: NvimFacade) -> NvimIO[A]:
        return (
            self.thunk(vim)
            .flat_map(self.f, fs=Just(self.fs))
            .with_stack(self.stack)
        )

    def _flat_map(self, f: Callable[[A], NvimIO[B]], ts: Eval[str], fs: Eval[str]) -> NvimIO[B]:
        def bs(vim: NvimFacade) -> NvimIO[B]:
            return BindSuspend(self.thunk, lambda a: self.f(a).flat_map(f, Just(ts), Just(fs)), ts, fs)
        return Suspend(bs, (ts & fs).map2('{}.{}'.format))


class Pure(Generic[A], NvimIO[A]):

    def __init__(self, value: A) -> None:
        super().__init__()
        self.value = value

    def lambda_str(self) -> Eval[str]:
        return Eval.later(lambda: f'Pure({self.value})')

    def _arg_desc(self) -> List[str]:
        return List(str(self.value))

    def step1(self, vim: NvimFacade) -> NvimIO[A]:
        return self

    def _flat_map(self, f: Callable[[A], NvimIO[B]], ts: Eval[str], fs: Eval[str]) -> NvimIO[B]:
        def g(vim: NvimFacade) -> NvimIO[B]:
            return f(self.value)
        return Suspend(g, (ts & fs).map2('{}.{}'.format))


class NvimIOError(Generic[A], NvimIO[A]):

    def __init__(self, error: str) -> None:
        self.error = error

    def _arg_desc(self) -> List[str]:
        return List(str(self.value))

    def lambda_str(self) -> Eval[str]:
        return Eval.later(lambda: f'NvimIOError({self.error})')

    def _flat_map(self, f: Callable[[A], NvimIO[B]], ts: Eval[str], fs: Eval[str]) -> NvimIO[B]:
        return self

    def step1(self, vim: NvimFacade) -> NvimIO[A]:
        return self


class NvimIOMonad(Monad[NvimIO]):

    def pure(self, a: A) -> NvimIO[A]:
        return NvimIO.pure(a)

    def flat_map(self, fa: NvimIO[A], f: Callable[[A], NvimIO[B]]) -> NvimIO[B]:
        return fa.flat_map(f)

    def map(self, fa: NvimIO[A], f: Callable[[A], B]) -> NvimIO[B]:
        return NvimIO.delay(lambda a: f(fa.run(a)))


class NvimIOState(Generic[S, A], StateT[NvimIO, S, A], tpe=NvimIO):

    @staticmethod
    def io(f: Callable[[NvimFacade], A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(NvimIO.delay(f))

    @staticmethod
    def delay(f: Callable[[NvimFacade], A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(NvimIO.delay(f))

    @staticmethod
    def suspend(f: Callable[[NvimFacade], NvimIO[A]]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(NvimIO.suspend(f))

    @staticmethod
    def from_io(io: IO[A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(NvimIO.wrap_either(lambda v: io.attempt))

    @staticmethod
    def from_id(st: State[S, A]) -> 'NvimIOState[S, A]':
        return st.transform_f(NvimIOState, lambda s: NvimIO.pure(s.value))

    @staticmethod
    def from_either(e: Either[str, A]) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(NvimIO.from_either(e))

    @staticmethod
    def failed(e: str) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(NvimIO.failed(e))

    @staticmethod
    def error(e: str) -> 'NvimIOState[S, A]':
        return NvimIOState.lift(NvimIO.error(e))


tcs(NvimIO, NvimIOState)  # type: ignore

NS = NvimIOState


class ToNvimStateIO(TypeClass):

    @abc.abstractproperty
    def nvim(self) -> NS:
        ...


class IdStateToNvimStateIO(ToNvimStateIO, tpe=State):

    @tc_prop
    def nvim(self, fa: State[S, A]) -> NS:
        return NvimIOState.from_id(fa)

__all__ = ('NvimIO', 'NvimIOState', 'NS')



import sys
import typing
from typing import Union, Any, Callable, Type, Optional, TypeVar, Generic, GenericMeta

import neovim
from neovim.api import Nvim

from amino import List
from amino.json import decode_json

import ribosome.nvim.components
from ribosome.nvim import NvimFacade
from ribosome.logging import nvim_logging, Logging
from ribosome.request.command import command
from ribosome.request.function import function
from ribosome.request.rpc import rpc_handlers_json
from ribosome.config import Config, PluginSettings


class NvimPluginBase(Logging):
    name: Optional[str] = None
    debug: bool = False

    @property
    def plugin_name(self) -> str:
        return type(self).name or 'ribosome'

    def __init__(self, nvim: Union[NvimFacade, neovim.Nvim]) -> None:
        self.vim = NvimFacade(nvim, self.plugin_name) if isinstance(nvim, neovim.Nvim) else nvim

    def setup_logging(self) -> None:
        self.file_log_handler = nvim_logging(self.vim)


NB = TypeVar('NB', bound=NvimPluginBase)
N = TypeVar('N', bound='NvimPlugin')
NM = TypeVar('NM', bound='NvimPluginMeta')


class NvimPluginMeta(GenericMeta):

    def __new__(
            cls: Type[NM],
            name: str,
            bases: tuple,
            namespace: dict,
            pname: str=None,
            prefix: str=None,
            debug: bool=False
    ) -> Type[N]:
        inst: Type[N] = super().__new__(cls, name, bases, namespace)
        if pname is not None:
            setup_plugin(inst, pname, prefix or pname, debug)
        return inst


class NvimPlugin(NvimPluginBase, metaclass=NvimPluginMeta):
    prefix: Optional[str] = None

    def quit(self) -> None:
        pass

    def rpc_handlers(self) -> List[str]:
        return rpc_handlers_json(type(self))

    def set_log_level(self, level: str) -> None:
        self.file_log_handler.setLevel(level)

    def append_python_path(self, path: str) -> None:
        sys.path.append(path)

    def show_python_path(self) -> typing.List[str]:
        return sys.path

    @neovim.autocmd('VimLeave', sync=True)
    def vim_leave(self) -> None:
        ribosome.nvim.components.shutdown = True
        self.quit()

    def send_message(self, data: str) -> None:
        return decode_json(data) / self.root.send


NSP = TypeVar('NSP', bound='NvimStatePlugin')
NSPM = TypeVar('NSPM', bound='NvimStatePluginMeta')


class NvimStatePluginMeta(NvimPluginMeta):

    def __new__(
            cls: Type[NSPM],
            name: str,
            bases: tuple,
            namespace: dict,
            pname: str=None,
            prefix: str=None,
            debug: bool=False
    ) -> Type[NSP]:
        inst: Type[NSP] = super().__new__(cls, name, bases, namespace, pname, prefix, debug)
        if inst.name is not None:
            setup_state_plugin(inst, inst.name, inst.prefix or inst.name, debug)
        return inst


class NvimStatePlugin(NvimPlugin, metaclass=NvimStatePluginMeta):

    def state_data(self) -> str:
        return self.state().data.json.value_or(lambda a: f'could not serialize state: {a}')

    def plug_command(self, plug_name: str, cmd_name: str, *args: str) -> None:
        self.state().plug_command(plug_name, cmd_name, args)


Settings = TypeVar('Settings', bound=PluginSettings)
D = TypeVar('D')
APM = TypeVar('APM', bound='AutoPluginMeta')
AP = TypeVar('AP', bound='AutoPlugin')


class AutoPluginMeta(NvimStatePluginMeta):

    def __new__(
            cls: Type[APM],
            name: str,
            bases: tuple,
            namespace: dict,
            pname: str=None,
            prefix: str=None,
            debug: bool=False,
            config: Config[Settings, D]=None,
    ) -> Type['AutoPlugin[Settings, D]']:
        return super().__new__(cls, name, bases, namespace, pname, prefix, debug)


class AutoPlugin(Generic[Settings, D], NvimStatePlugin, metaclass=AutoPluginMeta):

    def __init__(self, nvim: Union[NvimFacade, neovim.Nvim], config: Config[Settings, D]) -> None:
        super().__init__(nvim)
        self.config = config


class Helpers(Logging):

    def __init__(self, cls: type, name: str, prefix: str) -> None:
        self.cls = cls
        self.name = name
        self.prefix = prefix

    def handler(self, name: str, dec: Callable[..., Callable], handler: Callable, *a: Any, **kw: Any) -> None:
        def wrap(self: Any, *a: Any, **kw: Any) -> Any:
            try:
                return handler(self, *a, **kw)
            except Exception as e:
                self.log.caught_exception_error(f'calling handler `{name}`', e)
        func = dec(*a, name=name, **kw)(wrap)
        setattr(self.cls, name, func)

    def name_handler(self, suf: str, dec: Callable[..., Callable], handler: Callable, *a: Any, **kw: Any) -> None:
        n = f'{self.name}_{suf}'
        self.handler(n, dec, handler, *a, **kw)

    def short_handler(self, suf: str, dec: Callable[..., Callable], handler: Callable, *a: Any, **kw: Any) -> None:
        n = f'{self.prefix}_{suf}'
        self.handler(n, dec, handler, *a, **kw)

    # def msg_cmd(self, suf: str, msg: type) -> None:
    #     self.short_handler(suf, msg_command, lambda: None, msg)

    # def msg_fun(self, suf: str, msg: type) -> None:
    #     self.short_handler(suf, msg_function, lambda: None, msg)

    # def json_msg_cmd(self, suf: str, msg: type) -> None:
    #     self.short_handler(suf, json_msg_command, lambda: None, msg)


def setup_plugin(cls: Type[NvimPlugin], name: str, prefix: str, debug: bool) -> None:
    help = Helpers(cls, name, prefix)
    cls.name = name
    cls.prefix = prefix
    cls.debug = debug
    # help.msg_cmd('show_log_info', ShowLogInfo)
    help.short_handler('log_level', command, cls.set_log_level)
    # help.msg_fun('mapping', Mapping)
    # help.name_handler('stage_1', command, cls.stage_1, sync=True)
    # help.name_handler('stage_2', command, cls.stage_2, sync=True)
    # help.name_handler('stage_3', command, cls.stage_3, sync=True)
    # help.name_handler('stage_4', command, cls.stage_4, sync=True)
    # help.name_handler('quit', command, cls.quit, sync=True)
    help.name_handler('rpc_handlers', function, cls.rpc_handlers, sync=True)
    help.name_handler('append_python_path', function, cls.append_python_path)
    help.name_handler('show_python_path', function, cls.show_python_path)
    help.name_handler('send', function, cls.send_message)


def setup_state_plugin(cls: Type[NSP], name: str, prefix: str, debug: bool) -> None:
    help = Helpers(cls, name, prefix)
    help.short_handler('state', function, cls.state_data)
    # help.json_msg_cmd('update_state', UpdateState)
    help.short_handler('plug', command, cls.plug_command)
    # if debug:
    #     help.name_handler('message_log', function, cls.message_log, sync=True)


def plugin_class_from_config(config: Config[Settings, D], cls: Type[AP], debug=bool) -> Type[AP]:
    class Plug(AutoPlugin, config=config, pname=config.name, prefix=config.prefix, debug=debug):
        def __init__(self, vim: Nvim) -> None:
            super().__init__(vim, config)
    return type(cls.__name__, (cls, Plug), {})

__all__ = ('NvimPlugin', 'NvimStatePlugin')



import abc
from typing import Callable, TypeVar, Generic, Union, Any, Optional

from amino import List, Nil, Maybe, Map
from amino.dat import Dat

from ribosome.nvim import NvimIO
from ribosome.request.handler.handler import RequestHandler, RequestHandlers
from ribosome.config.settings import PluginSettings
from ribosome.dispatch.run import DispatchJob

A = TypeVar('A', contravariant=True)
B = TypeVar('B')
Settings = TypeVar('Settings', bound=PluginSettings)
D = TypeVar('D', bound='SimpleData')


class Config(Generic[Settings, D], Dat['Config[Settings, D]']):

    @staticmethod
    def from_opt(data: Map) -> 'Config[Settings, D]':
        return Config.cons(data.lift('name') | 'no name in json', data.lift('prefix') | None)

    @staticmethod
    def cons(
            name: str,
            prefix: Optional[str]=None,
            components: Map[str, Union[str, type]]=Map(),
            state_ctor: Optional[Callable[['Config'], D]]=None,
            settings: Optional[Settings]=None,
            request_handlers: List[RequestHandler]=Nil,
            core_components: List[str]=Nil,
            default_components: List[str]=Nil,
            request_dispatcher: Optional[Callable[[DispatchJob], NvimIO[Any]]]=None,
    ) -> 'Config[Settings, D]':
        return Config(
            name,
            prefix or name,
            components,
            state_ctor or SimpleData,
            settings or PluginSettings(name=name),
            RequestHandlers.cons(*request_handlers),
            core_components,
            default_components,
            Maybe.optional(request_dispatcher),
        )

    def __init__(
            self,
            name: str,
            prefix: str,
            components: Map[str, Union[str, type]],
            state_ctor: Callable[['Config'], D],
            settings: Settings,
            request_handlers: RequestHandlers,
            core_components: List[str],
            default_components: List[str],
            request_dispatcher: Maybe[Callable[[DispatchJob], NvimIO[Any]]],
    ) -> None:
        self.name = name
        self.prefix = prefix
        self.components = components
        self.state_ctor = state_ctor
        self.settings = settings
        self.request_handlers = request_handlers
        self.core_components = core_components
        self.default_components = default_components
        self.request_dispatcher = request_dispatcher

    def _arg_desc(self) -> List[str]:
        return List(str(self.components), str(self.settings), str(self.request_handlers))

    def state(self) -> D:
        return self.state_ctor(self)

    def vim_cmd_name(self, handler: RequestHandler) -> str:
        return handler.vim_cmd_name(self.name, self.prefix)


class Data:

    def __init__(self, config: Config) -> None:
        self.config = config

    @property
    def _str_extra(self) -> List[Any]:
        return List(self.config)

    @property
    def settings(self) -> Settings:
        return self.config.settings


class SimpleData(Dat['SimpleData'], Data):

    def __init__(self, config: Config) -> None:
        super().__init__(config)


__all__ = ('Config', 'Data', 'SimpleData')



import abc
import inspect
from types import FunctionType
from typing import TypeVar, Callable, Union, Any, Dict, Generator

from amino import Map, Maybe, Lists, List, Either, Just, Nothing, do, Boolean, _, L, __
from amino.util.string import camelcaseify, ToStr
from amino.list import Nil
from amino.dat import Dat

from ribosome.nvim import NvimFacade, NvimIO
from ribosome.logging import ribo_log

A = TypeVar('A')


def rpc_cmd_opt(key: str, value: Union[str, int]) -> Maybe[str]:
    return (
        Just(f'-nargs={value}')
        if key == 'nargs' else
        Just('-bang')
        if key == 'bang' else
        Nothing
    )


def rpc_cmd_arg(key: str, value: Union[str, int]) -> Maybe[str]:
    return (
        Just(f'[<f-args>]')
        if key == 'nargs' else
        Just('<q-bang> == "!"')
        if key == 'bang' else
        Nothing
    )


def rpc_autocmd_opt(key: str, value: Union[str, int]) -> Maybe[str]:
    return (
        Just(str(value))
        if key == 'pattern' else
        Nothing
    )


RHS = TypeVar('RHS', bound='RpcHandlerSpec')


def from_bool(a: Union[int, bool, Boolean]) -> int:
    return a.to_int if isinstance(a, Boolean) else int(a) if isinstance(a, bool) else a


class RpcHandlerSpec(Dat['RpcHandlerSpec']):

    @staticmethod
    def cons(Ctor: type,
             sync: Union[int, bool, Boolean],
             name: str,
             opts: dict,
             method: str,
             prefix: bool) -> 'RpcHandlerSpec':
        return Ctor(from_bool(sync), name, opts, method, Boolean(prefix))

    @staticmethod
    @do(Maybe['RpcHandlerSpec'])
    def from_spec(spec: Dict[str, Any], method: str, prefix: bool) -> Generator:
        m = Map(spec)
        ctors = Map(command=RpcCommandSpec, function=RpcFunctionSpec, autocmd=RpcAutocommandSpec)
        tpe = yield m.lift('type')
        Ctor = yield ctors.lift(tpe)
        sync, name, opts = yield m.get_all('sync', 'name', 'opts')
        yield Just(RpcHandlerSpec.cons(Ctor, sync, name, opts, method, prefix))

    @staticmethod
    def cmd(sync: int, name: str, opts: dict) -> 'RpcHandlerSpec':
        method = f'command:{name}'
        return RpcHandlerSpec.cons(RpcCommandSpec, sync, name, opts, method, True)

    @staticmethod
    def fun(sync: int, name: str, opts: dict) -> 'RpcHandlerSpec':
        method = f'function:{name}'
        return RpcHandlerSpec.cons(RpcFunctionSpec, sync, name, opts, method, True)

    @staticmethod
    def autocmd(sync: int, name: str, opts: dict) -> 'RpcHandlerSpec':
        method = f'autocmd:{name}'
        return RpcHandlerSpec.cons(RpcAutocommandSpec, sync, name, opts, method, True)

    @staticmethod
    @do('Maybe[RpcHandlerSpec]')
    def decode(data: dict) -> Generator:
        m = Map(data)
        method, prefix = yield m.get_all('method', 'prefix')
        yield RpcHandlerSpec.from_spec(m, method, prefix)

    @abc.abstractproperty
    def tpe(self) -> str:
        ...

    @abc.abstractproperty
    def rpc_opts_pre(self) -> List[str]:
        ...

    @abc.abstractproperty
    def rpc_opts_post(self) -> List[str]:
        ...

    @abc.abstractproperty
    def rpc_args(self) -> List[str]:
        ...

    @abc.abstractproperty
    def def_cmd(self) -> str:
        ...

    def __init__(self, sync: Boolean, name: str, opts: dict, method: str, prefix: Boolean) -> None:
        self.sync = sync
        self.name = name
        self.opts = opts
        self.method = method
        self.prefix = prefix

    def _arg_desc(self) -> List[str]:
        return List(self.name, self.tpe, str(self.sync), str(self.opts), self.method, str(self.prefix))

    @property
    def rpc_function(self) -> str:
        return 'rpcrequest' if self.sync else 'rpcnotify'

    def rpc_method(self, prefix: str) -> str:
        pre = f'{prefix}:' if self.prefix else ''
        return f'{pre}{self.method}'

    @property
    def encode(self) -> dict:
        return dict(
            type=self.tpe,
            sync=self.sync,
            name=self.name,
            opts=self.opts,
            method=self.method,
            prefix=int(self.prefix.value),
        )

    @abc.abstractproperty
    def undef_cmdline(self) -> str:
        ...


class RpcCommandArgs:

    def __init__(self, data: Map) -> None:
        self.data = data

    @property
    def tokens(self) -> List[str]:
        return self.nargs + self.bang

    @property
    def nargs(self) -> List[str]:
        return self.arg('nargs', lambda a: List('[<f-args>]'))

    @property
    def bang(self) -> List[str]:
        return self.arg('bang', lambda a: List("<q-bang> == '!'"))

    def arg(self, name: str, result: Callable[[str], List[str]]) -> List[str]:
        return self.data.lift(name) / result | Nil


class RpcCommandSpec(RpcHandlerSpec):

    @property
    def tpe(self) -> str:
        return 'command'

    @property
    def rpc_opts_pre(self) -> List[str]:
        return Map(self.opts).map2(rpc_cmd_opt).join

    @property
    def rpc_opts_post(self) -> List[str]:
        return Nil

    @property
    def rpc_args(self) -> List[str]:
        help = RpcCommandArgs(Map(self.opts))
        return help.tokens

    @property
    def def_cmd(self) -> str:
        return 'command!'

    @property
    def undef_cmdline(self) -> str:
        return f'delcommand {self.name}'


class RpcFunctionSpec(RpcHandlerSpec):

    @property
    def tpe(self) -> str:
        return 'function'

    @property
    def rpc_opts_pre(self) -> List[str]:
        return Nil

    @property
    def rpc_opts_post(self) -> List[str]:
        return Nil

    @property
    def rpc_args(self) -> List[str]:
        return List('a:000')

    @property
    def def_cmd(self) -> str:
        return 'function!'

    @property
    def undef_cmdline(self) -> str:
        return f'delfunction {self.name}'


class RpcAutocommandSpec(RpcHandlerSpec):

    @property
    def tpe(self) -> str:
        return 'autocmd'

    @property
    def rpc_opts_pre(self) -> List[str]:
        return List()

    @property
    def rpc_opts_post(self) -> List[str]:
        defaults = Map(pattern='*')
        return Map(defaults ** self.opts).map2(rpc_autocmd_opt).join

    @property
    def rpc_args(self) -> List[str]:
        return Nil

    @property
    def def_cmd(self) -> str:
        return 'autocmd'

    @property
    def undef_cmdline(self) -> str:
        return f'autocmd! {self.name}'


def handler(method_name: str, fun: FunctionType) -> Maybe[RpcHandlerSpec]:
    def get(name: str) -> Maybe[A]:
        return Maybe.getattr(fun, name)
    return (
        List('_nvim_rpc_spec', '_nvim_rpc_method_name', '_nvim_prefix_plugin_path')
        .traverse(get, Maybe)
        .flat_map3(RpcHandlerSpec.from_spec)
    )


class RpcHandlerFunction(ToStr):

    def __init__(self, func: Callable, spec: RpcHandlerSpec) -> None:
        self.func = func
        self.spec = spec

    def _arg_desc(self) -> List[str]:
        return List(str(self.spec))


def handler_function(method_name: str, fun: FunctionType) -> Maybe[RpcHandlerFunction]:
    return handler(method_name, fun) / L(RpcHandlerFunction)(fun, _)


def handler_function1(method_name: str, fun: FunctionType, name: str, prefix: str) -> Maybe[RpcHandlerFunction]:
    return Maybe.getattr(fun, 'spec') / __.spec(name, prefix) / L(RpcHandlerFunction)(fun, _)


def register_handler_args(host: str, spec: RpcHandlerSpec, plugin_file: str) -> List[str]:
    fun_prefix = camelcaseify(spec.tpe)
    return List(f'remote#define#{fun_prefix}OnHost', host, spec.rpc_method, spec.sync, spec.name, spec.opts)


def define_handler_native(vim: NvimFacade, host: str, spec: RpcHandlerSpec, plugin_file: str
                          ) -> Either[Exception, None]:
    ribo_log.debug1(lambda: f'defining {spec} on {host}')
    args = register_handler_args(host, spec, plugin_file)
    return vim.call(*args)


class DefinedHandler(Dat['DefinedHandler']):

    def __init__(self, spec: RpcHandlerSpec, channel: int) -> None:
        self.spec = spec
        self.channel = channel


def quote(a: str) -> str:
    return f"'{a}'"


def define_handler_cmd(rhs_f: Callable[[str], str], channel: int, spec: RpcHandlerSpec, plugin_file: str) -> List[str]:
    ribo_log.debug1(lambda: f'defining {spec} on channel {channel}')
    fun = spec.rpc_function
    method = spec.rpc_method(plugin_file)
    args = spec.rpc_args.cons(quote(method)).join_comma
    rpc_call = f'{fun}({channel}, {args})'
    rhs = rhs_f(rpc_call)
    return List(spec.def_cmd) + spec.rpc_opts_pre + List(spec.name) + spec.rpc_opts_post + List(rhs)


def define_handler_io(rhs: Callable[[str], str], channel: int, spec: RpcHandlerSpec, plugin_file: str
                      ) -> NvimIO[DefinedHandler]:
    tokens = define_handler_cmd(rhs, channel, spec, plugin_file)
    return NvimIO.cmd_sync(tokens.join_tokens).replace(DefinedHandler(spec, channel))


def define_function(channel: int, spec: RpcHandlerSpec, plugin_file: str) -> NvimIO[DefinedHandler]:
    return define_handler_io(lambda call: f'(...)\nreturn {call}\nendfunction', channel, spec, plugin_file)


def define_command(channel: int, spec: RpcHandlerSpec, plugin_file: str) -> NvimIO[DefinedHandler]:
    return define_handler_io(lambda a: f'call {a}', channel, spec, plugin_file)


@do(NvimIO[DefinedHandler])
def define_autocmd(channel: int, spec: RpcHandlerSpec, plugin_name: str, plugin_file: str) -> Generator:
    yield NvimIO.cmd_sync(f'augroup {plugin_name}')
    result = yield define_handler_io(lambda a: f'call {a}', channel, spec, plugin_file)
    yield NvimIO.cmd_sync(f'augroup end')
    yield NvimIO.pure(result)


def define_handler(channel: int, spec: RpcHandlerSpec, plugin_name: str, plugin_file: str) -> NvimIO[DefinedHandler]:
    if spec.tpe == 'function':
        return define_function(channel, spec, plugin_file)
    elif spec.tpe == 'command':
        return define_command(channel, spec, plugin_file)
    elif spec.tpe == 'autocmd':
        return define_autocmd(channel, spec, plugin_name, plugin_file)
    else:
        return NvimIO.failed(f'invalid type for {spec}')


@do(NvimIO[List[DefinedHandler]])
def define_handlers(specs: List[RpcHandlerSpec], plugin_name: str, plugin_file: str) -> Generator:
    channel = yield NvimIO.delay(_.channel_id)
    yield specs.traverse(L(define_handler)(channel, _, plugin_name, plugin_file), NvimIO)


def rpc_handlers(plugin_class: type) -> List[RpcHandlerSpec]:
    return Lists.wrap(inspect.getmembers(plugin_class)).flat_map2(handler)


def rpc_handlers_json(plugin_class: type) -> List[str]:
        return list(rpc_handlers(plugin_class) / _.encode)


def rpc_handler_functions(plugin: Any) -> List[RpcHandlerFunction]:
    funs = inspect.getmembers(plugin)
    return (
        Lists.wrap(funs).flat_map2(handler_function) +
        Lists.wrap(funs).flat_map2(L(handler_function1)(_, _, plugin.name, plugin.prefix))
    )

__all__ = ('RpcHandlerSpec', 'handler', 'define_handler', 'define_handlers', 'rpc_handlers', 'rpc_handlers_json',
           'rpc_handler_functions')



import inspect
import logging
from typing import TypeVar, Generic, Type, Any, Callable
from threading import Lock

import toolz

from amino import Map, List, Boolean, Nil, Either, _, Lists, Maybe, Nothing
from amino.dat import Dat
from amino.func import flip

from ribosome.trans.message_base import Message, Sendable, Envelope
from ribosome.dispatch.component import Component
from ribosome.nvim.io import NvimIOState, NS
from ribosome.dispatch.data import DispatchResult, DIO
from ribosome.trans.queue import PrioQueue
from ribosome.trans.legacy import Handler
from ribosome.trans.handler import TransHandler, TransComplete

D = TypeVar('D')
TransState = NvimIOState[D, DispatchResult]


class Handlers(Dat['Handlers']):

    def __init__(self, prio: int, handlers: Map[type, Handler]) -> None:
        self.prio = prio
        self.handlers = handlers

    def handler(self, msg):
        return self.handlers.get(type(msg))


def handlers(cls: Type['MachineBase']) -> List[Handler]:
    return Lists.wrap(inspect.getmembers(cls, Boolean.is_a(TransHandler))) / _[1]


def message_handlers(handlers: List[Handler]) -> Map[float, Handlers]:
    def create(prio, h):
        h = List.wrap(h).apzip(_.message).map2(flip)
        return prio, Handlers(prio, Map(h))
    return Map(toolz.groupby(_.prio, handlers)).map(create)


class ComponentState(Dat['ComponentState']):

    @staticmethod
    def cons(comp: Component) -> 'ComponentState':
        hs = message_handlers(handlers(comp))
        return ComponentState(comp, hs)

    def __init__(self, component: Component, handlers: Map[float, Handlers]) -> None:
        self.component = component
        self.handlers = handlers

    @property
    def name(self) -> str:
        return self.component.name


class Components(Dat['Components']):

    def __init__(self, all: List[ComponentState]) -> None:
        self.all = all

    def by_name(self, name: str) -> Either[str, ComponentState]:
        return self.all.find(_.name == name).to_either(f'no component named {name}')


class DispatchConfig(Dat['DispatchConfig']):

    @staticmethod
    def cons(
            config: 'ribosome.config.Config',
            io_executor: Callable[[DIO], NS['PluginState[D]', TransComplete]]=None,
    ) -> 'DispatchConfig':
        return DispatchConfig(config, Maybe.optional(io_executor))

    def __init__(
            self,
            config: 'ribosome.config.Config',
            io_executor: Maybe[Callable[[DIO], NS['PluginState[D]', TransComplete]]]
    ) -> None:
        self.config = config
        self.io_executor = io_executor


class PluginState(Generic[D], Dat['PluginState']):

    @staticmethod
    def cons(
            dispatch_config: DispatchConfig,
            data: D,
            plugin: Any,
            components: List[Component],
            messages: PrioQueue[Message],
            message_log: List[Message]=Nil,
            log_handler: Maybe[logging.Handler]=Nothing,
    ) -> 'PluginState':
        component_state = Components(components / ComponentState.cons)
        return PluginState(dispatch_config, data, plugin, component_state, messages, message_log, log_handler)

    def __init__(
            self,
            dispatch_config: DispatchConfig,
            data: D,
            plugin: Any,
            components: List[ComponentState],
            messages: PrioQueue[Message],
            message_log: List[Message],
            log_handler: Maybe[logging.Handler],
    ) -> None:
        self.dispatch_config = dispatch_config
        self.data = data
        self.plugin = plugin
        self.components = components
        self.messages = messages
        self.message_log = message_log
        self.log_handler = log_handler

    def enqueue(self, messages: List[Sendable]) -> 'PluginState[D]':
        envelopes = messages / Envelope.from_sendable
        return self.copy(messages=envelopes.fold_left(self.messages)(lambda q, m: q.put(m, m.prio)))

    def update(self, data: D) -> 'PluginState[D]':
        return self.copy(data=data)

    def log_messages(self, msgs: List[Message]) -> 'PluginState[D]':
        return self.append.message_log(msgs)

    def log_message(self, msg: Message) -> 'PluginState[D]':
        return self.log_messages(List(msg))

    @property
    def has_messages(self) -> Boolean:
        return not self.messages.empty

    @property
    def unwrapped_messages(self) -> List[Message]:
        return self.messages.items / _[1] / _.msg

    @property
    def config(self) -> 'ribosome.config.Config':
        return self.dispatch_config.config

    @property
    def name(self) -> str:
        return self.config.name

    def component(self, name: str) -> Either[str, ComponentState]:
        return self.components.by_name(name)


class PluginStateHolder(Generic[D], Dat['PluginStateHolder']):

    @staticmethod
    def cons(state: PluginState[D], log_handler: logging.Handler=None) -> 'PluginStateHolder':
        return PluginStateHolder(state, Lock(), Maybe.check(log_handler))

    def __init__(self, state: PluginState[D], lock: Lock, log_handler: Maybe[logging.Handler]=Nothing) -> None:
        self.state = state
        self.lock = lock
        self.log_handler = log_handler

    def update(self, state: PluginState[D]) -> None:
        self.state = state

    def acquire(self) -> None:
        self.lock.acquire(timeout=10.0)

    def release(self) -> None:
        self.lock.release()

    @property
    def has_messages(self) -> Boolean:
        return self.state.has_messages

    @property
    def message_queue(self) -> PrioQueue[Message]:
        return self.state.messages

    @property
    def messages(self) -> List[Message]:
        return self.state.unwrapped_messages


__all__ = ('ComponentState', 'PluginState', 'PluginStateHolder', 'Components')



from typing import Callable, Any, Union, List as TList

import neovim

from amino import List, Map, L, __, _
from amino.util.string import snake_case

from ribosome.logging import log
from ribosome.unite.data import UniteSyntax


def _convert_candidate(c: Union[str, dict]) -> dict:
    return c if isinstance(c, dict) else {'word': c}


def mk_unite_candidates(Unite: type) -> Callable[[str], Callable[[Callable], Callable]]:
    def decorator(name: str) -> Callable:
        handler = getattr(Unite, '{}_candidates'.format(name))
        def uc_wrap(f: Callable) -> Callable:
            @neovim.function(handler, sync=True)
            def f_wrap(self: Any, args: list) -> List[dict]:
                return f(self, args) / _convert_candidate
            return f_wrap
        return uc_wrap
    return decorator


def _unite_word(args: TList[TList[dict]], key: str) -> str:
    first = List.wrap(args).lift(0)
    candidates = (
        first.to_list.join
        if first.exists(lambda a: isinstance(a, list))
        else first
    )
    return candidates / Map // __.get(key)


def mk_unite_action(Unite: type) -> Callable[[str, str], Callable[[Callable], Callable]]:
    def decorator(name: str, key: str='word') -> Callable[[Callable], Callable]:
        handler = getattr(Unite, name)
        def ua_wrap(f: Callable) -> Callable:
            @neovim.function(handler)
            def f_wrap(self: Any, args: list) -> None:
                _unite_word(args, key) / L(f)(self, _) % self.state().send
            return f_wrap
        return ua_wrap
    return decorator


def _unite_function(name: str, msg: type, param_cb: Callable) -> Callable:
    @neovim.function(name, sync=True)
    def _unite_dispatcher(self: Any, args: list) -> None:
        params = param_cb(List.wrap(args))
        params / UniteSyntax % self.state().send
    return _unite_dispatcher


def unite_plugin(pname: str) -> Callable[[type], type]:
    sname = snake_case(pname)
    syntax_name = '_{}_unite_syntax'.format(sname)
    def decorator(cls: type) -> type:
        def set_fun(fname: str, msg: type, cb: Callable) -> None:
            try:
                setattr(cls, fname, _unite_function(fname, msg, cb))
            except Exception as e:
                log.caught_exception_error(f'calling unite function {fname}', e)
        syntax_cb = lambda args: args.tail // _.head / __['source'] / __['pname']
        set_fun(syntax_name, UniteSyntax, syntax_cb)
        return cls
    return decorator

__all__ = ('mk_unite_candidates', 'mk_unite_action', 'unite_plugin')



from ribosome.trans.effect import (TransEffect, TransEffectMaybe, TransEffectEither, TransEffectStateT, TransEffectIO,
                                   TransEffectNvimIO, TransEffectCoro, TransEffectSingleMessage, TransEffectMessages,
                                   TransEffectUnit, TransEffectResult, TransEffectGatherIOs, TransEffectDo,
                                   TransEffectGatherSubprocs)
from ribosome.trans.main import FreeTransCons, MessageTransCons


class TransApi:
    m: TransEffect = TransEffectMaybe()
    e: TransEffect = TransEffectEither()
    st: TransEffect = TransEffectStateT()
    io: TransEffect = TransEffectIO()
    gather_ios: TransEffect = TransEffectGatherIOs()
    gather_subprocs: TransEffect = TransEffectGatherSubprocs()
    nio: TransEffect = TransEffectNvimIO()
    coro: TransEffect = TransEffectCoro()
    single: TransEffect = TransEffectSingleMessage()
    strict: TransEffect = TransEffectMessages()
    none: TransEffect = TransEffectUnit()
    result: TransEffect = TransEffectResult()
    do: TransEffect = TransEffectDo()
    free: FreeTransCons = FreeTransCons()
    msg: MessageTransCons = MessageTransCons()


trans = TransApi()

__all__ = ('TransApi', 'trans')



import time
from typing import TypeVar, Callable

import amino
from amino import __, Maybe, Boolean, _, Nil, L, Try
from amino.do import do, Do

from ribosome.trans.message_base import Message, Sendable, Envelope
from ribosome.logging import ribo_log
from ribosome.plugin_state import PluginState, ComponentState, Components, TransState
from ribosome.nvim.io import NvimIOState
from ribosome.dispatch.data import DispatchResult, DispatchError, DispatchOutputAggregate
from ribosome.dispatch.transform import validate_trans_action
from ribosome.dispatch.component import Component
from ribosome.trans.handler import MessageTransHandler
from ribosome.trans.legacy import TransitionFailed

A = TypeVar('A')
D = TypeVar('D')
M = TypeVar('M', bound=Sendable)
NP = TypeVar('NP')


def resolve_handler(component: ComponentState, msg: Message, prio: float=None) -> Maybe[Callable]:
    f = __.handler(msg)
    return (
        component.handlers.v.find_map(f)
        if prio is None else
        (component.handlers.get(prio) // f)
    )


def internal(msg: Message) -> TransState:
    return DispatchResult.unit_nio


def format_report(msg: Message, dur: float, name: str) -> str:
    return '{} took {:.4f}s for {} to process'.format(msg, dur, name)


def check_time(start_time: int, msg: Message, name: str) -> None:
    dur = time.time() - start_time
    ribo_log.debug1(format_report, msg, dur, name)
    # if dur > self._min_report_time:
    #     self._reports = self._reports.cat((msg, dur))


def handler_exception(e: Exception, desc: str) -> NvimIOState[D, DispatchResult]:
    if amino.development:
        err = f'transitioning {desc}'
        ribo_log.caught_exception(err, e)
        raise TransitionFailed(str(e)) from e
    return NvimIOState.pure(DispatchResult(DispatchError.cons(e), Nil))


def try_handler(handler: Callable[[M], A], msg: M, desc: str) -> NvimIOState[D, DispatchResult]:
    return Try(handler, msg).map(validate_trans_action).value_or(L(handler_exception)(_, desc))


def execute_component_handler(component: Component, handler: MessageTransHandler, msg: M
                              ) -> NvimIOState[D, DispatchResult]:
    result = try_handler(L(handler.run)(component, _), msg, component.name)
    # check_time(job.start_time, msg, component.name)
    return result


def process(component: ComponentState, msg: Message, prio: float) -> TransState:
    def execute(handler: Callable) -> TransState:
        ribo_log.debug(f'handling {msg} in {component.name}')
        return execute_component_handler(component.component, handler, msg)
    return resolve_handler(component, msg, prio) / execute | (lambda: internal(msg))


def send_msg(components: Components, msg: Message, prio: float) -> TransState:
    return (
        components.all
        .traverse(lambda comp: process(comp, msg, prio), NvimIOState) /
        DispatchOutputAggregate /
        L(DispatchResult)(_, Nil)
    )


def send_envelope(components: Components, env: Envelope, prio: float) -> TransState:
    msg = env.msg
    def to_comp(name: str) -> TransState:
        return components.by_name(name) / L(process)(_, msg, prio) | (lambda: DispatchResult.unit_nio)
    return env.recipient / to_comp | (lambda: send_msg(components, msg, prio))


@do(TransState)
def send_message1(components: Components, msg: M, prio: float) -> Do:
    ''' send **msg** to all submachines, passing the transformed data from each machine to the next and accumulating
    published messages.
    '''
    send = Boolean.isinstance(msg, Envelope).cata(send_envelope, send_msg)
    yield send(components, msg, prio)


def transform_data_state(st: NvimIOState[D, DispatchResult]) -> NvimIOState[PluginState[D], DispatchResult]:
    return st.transform_s(_.data, lambda r, s: r.copy(data=s))


@do(NvimIOState[PluginState[D], DispatchResult])
def send_message(msg: M, prio: float=None) -> Do:
    yield NvimIOState.modify(__.log_message(msg.msg))
    components = yield NvimIOState.inspect(_.components)
    yield transform_data_state(send_message1(components, msg, prio))


__all__ = ('send_message',)



from kallikrein import Expectation, kf
from kallikrein.matchers.either import be_right
from kallikrein.matchers import contain

from amino import Right, Either
from amino.test import temp_dir

from ribosome import NvimPlugin, function

from integration._support.base import IntegrationSpecBase

class_name = 'ZeeKlass'


class DefaultHandlerSpecPlugin(NvimPlugin, pname='plug'):

    @function()
    def test_path(self) -> str:
        return Either.import_name('pkg', class_name).map(lambda a: a.__name__).value_or('failed')


class DefaultHandlerSpec(IntegrationSpecBase):
    '''
    append a directory to the plugin's `sys.path` $append_path
    '''

    @property
    def _prefix(self) -> str:
        return 'DH'

    @property
    def plugin_class(self) -> Either[str, type]:
        return Right(DefaultHandlerSpecPlugin)

    def append_path(self) -> Expectation:
        pkg = temp_dir('default_handler', 'pp', 'pkg')
        pp = pkg.parent
        file = pkg / '__init__.py'
        file.write_text(f'class {class_name}: pass')
        self.vim.call('PlugAppendPythonPath', str(pp))
        return (
            kf(self.vim.call, 'TestPath').must(be_right(class_name)) &
            kf(self.vim.call, 'PlugShowPythonPath').must(be_right(contain(str(pp))))
        )

__all__ = ('DefaultHandlerSpec',)



from typing import TypeVar, Generic, Any, Callable
from types import SimpleNamespace

from amino import Either, List, Id, Maybe, Nil
from amino.state import StateT
from amino.util.string import ToStr
from amino.algebra import AlgebraMeta
from amino.dat import ADT, ADTMeta
from amino.tc.base import Implicits, ImplicitsMeta
from amino.tc.monad import Monad

from ribosome.trans.message_base import Sendable, Message
from ribosome.trans.messages import Error

A = TypeVar('A')
B = TypeVar('B')
D = TypeVar('D')
M = TypeVar('M', bound=Message)
R = TypeVar('R')
N = TypeVar('N')
O = TypeVar('O')
G = TypeVar('G')
I = TypeVar('I')


class TransMMeta(ADTMeta, ImplicitsMeta):

    def __new__(cls, name: str, bases: List[type], namespace: SimpleNamespace, **kw: Any) -> None:
        return super().__new__(cls, name, bases, namespace, **kw)


class TransM(Generic[A], ADT['TransM'], Implicits, implicits=True, auto=True, base=True, metaclass=TransMMeta):
    pass


class TransMPure(TransM):

    def __init__(self, handler: Any) -> None:
        self.handler = handler


class TransMBind(TransM):

    def __init__(self, fa: TransM[A], f: Callable[[A], TransM[B]]) -> None:
        super().__init__()
        self.fa = fa
        self.f = f


class Monad_TransM(Monad, tpe=TransM):

    def pure(self, a: A) -> TransM[A]:
        from ribosome.trans.api import trans
        @trans.free.result()
        def pure() -> None:
            return a
        return TransMPure(pure)

    def flat_map(self, fa: TransM, f: Callable[[A], TransM[B]]) -> None:
        return TransMBind(fa, f)


class TransAction(ADT['TransAction'], base=True):

    def __init__(self, messages: List[Sendable]) -> None:
        self.messages = messages


class Transit(Generic[D], TransAction):

    def __init__(self, trans: StateT[Id, D, 'Propagate']) -> None:
        super().__init__(List())
        self.trans = trans

    def _arg_desc(self) -> List[str]:
        return List(str(self.trans))


class Propagate(TransAction):

    @staticmethod
    def one(msg: Sendable) -> TransAction:
        return Propagate(List(msg))

    @staticmethod
    def maybe(msg: Maybe[Sendable]) -> TransAction:
        return Propagate(msg.to_list)

    @staticmethod
    def from_either(res: Either[str, TransAction]) -> TransAction:
        return res.value_or(lambda a: Propagate.one(Error(a)))

    def _arg_desc(self) -> List[str]:
        return self.messages / str


class TransUnit(TransAction):

    def __init__(self) -> None:
        super().__init__(Nil)

    def _arg_desc(self) -> List[str]:
        return List()


class TransResult(TransAction):

    def __init__(self, data: Any) -> None:
        self.data = data
        super().__init__(Nil)


class TransFailure(TransAction):

    def __init__(self, message: str) -> None:
        super().__init__(List())
        self.message = message

    def _arg_desc(self) -> List[str]:
        return List(self.message)


class TransIO(Generic[I], TransAction):

    def __init__(self, io: I) -> None:
        self.io = io


class TransDo(TransAction):

    def __init__(self, action: TransM) -> None:
        self.action = action


class TransStep(Generic[R], ToStr, metaclass=AlgebraMeta, base=True):

    def __init__(self, data: R) -> None:
        self.data = data

    @property
    def strict(self) -> bool:
        return isinstance(self, Strict)

    @property
    def error(self) -> bool:
        return isinstance(self, TransEffectError)

    def _arg_desc(self) -> List[str]:
        return List(str(self.data))


class Strict(Generic[R], TransStep[R]):
    pass


class Lift(Generic[R], TransStep[R]):
    pass


class TransEffectError(TransStep[str]):
    pass


__all__ = ('TransAction', 'Transit', 'Propagate', 'TransUnit', 'TransResult', 'TransFailure', 'TransStep', 'Strict',
           'Lift', 'TransEffectError')



import functools

import neovim

from ribosome.request.legacy import LegacyRequestHandler
from ribosome.request.handler.method import RpcMethod, CmdMethod


class Command(LegacyRequestHandler):

    @property
    def desc(self):
        return 'command'

    @property
    def neovim_cmd(self):
        @neovim.command(self.vim_name, nargs=self.nargs, **self.kw)
        @functools.wraps(self._fun)
        def neovim_cmd_wrapper(obj, *rpc_args):
            return self.dispatch(obj, rpc_args)
        return neovim_cmd_wrapper

    @property
    def method(self) -> RpcMethod:
        return CmdMethod()


def command(**kw):
    def neovim_cmd_decorator(fun):
        handler = Command(fun, **kw)
        ncmd = handler.neovim_cmd
        ncmd.spec = handler.spec
        return ncmd
    return neovim_cmd_decorator


__all__ = ('command',)



from typing import Callable, Any, Optional
from contextlib import contextmanager

from amino import List, Map, Boolean, Maybe, Just
from amino.util.fun import format_funcall

from ribosome.nvim import Buffer, Tab, Window
from ribosome import NvimFacade
from ribosome.nvim.components import Options, Vars


class MockOptions(Options):

    def set(self, *a, **kw):
        pass


class MockVars(Vars):

    def __init__(self, vim, vars: dict) -> None:
        super().__init__(vim)
        self.vars = vars
        self.prefix = vim.prefix

    def _get(self, name: str) -> Optional[Any]:
        return self.vars.get(name)

    def set(self, name, value):
        self.vars[name] = value


class MockNvimComponent(object):

    def __init__(self, prefix, vars: dict) -> None:
        self.prefix = prefix
        self._vars = MockVars(self, vars)
        self._options = MockOptions(self)

    @property
    def window(self):
        return MockWindow(self.vim, None, self.prefix)

    @property
    def buffer(self):
        return MockBuffer(self.vim, None, self.prefix)

    @property
    def tab(self):
        return MockTab(self.vim, None, self.prefix)


class MockTab(MockNvimComponent, Tab):
    pass


class MockWindow(MockNvimComponent, Window):
    pass


class MockBuffer(MockNvimComponent, Buffer):
    pass


class MockNvim:

    def __init__(self, responses: Callable[[str], Any]) -> None:
        self.responses = responses

    @property
    def types(self) -> Map[str, type]:
        return Map()

    def command(self, cmdline: str, **kw: Any) -> Any:
        return self.responses(cmdline).get_or_fail(f'no response recorded for command {cmdline}')

    def call(self, name: str, *a: Any, **kw: Any) -> Any:
        rep = format_funcall(name, a, kw)
        return self.responses(name).get_or_fail(f'no response recorded for call {rep}')


class MockNvimFacade(MockNvimComponent, NvimFacade):

    def __init__(self, prefix: str='ribosome', vars: dict=dict(),
                 responses: Callable[[str], Maybe[Any]]=lambda a: Just(0)) -> None:
        MockNvimComponent.__init__(self, prefix, vars)
        self.vim = MockNvim(responses)
        self.target = self.vim

    @property
    def windows(self):
        return List()

    def switch_root(self, root):
        pass

    def async(self, f: Callable[['NvimFacade'], Any], *a, **kw):
        return f(self, *a, **kw)

    @contextmanager
    def main_event_loop(self):
        yield None

    def reload_windows(self):
        pass

    def function_exists(self, name: str) -> Boolean:
        return self.vim.responses(name).present

__all__ = ('MockNvimFacade',)



from typing import Callable, TypeVar, Type

from amino import Lists
from ribosome.trans.effect import (TransEffect, TransEffectUnit, TransEffectSingleMessage, TransEffectMessages,
                                   TransEffectResult, TransEffectDo)
from ribosome.trans.handler import FreeTransHandler, MessageTransHandler
from ribosome.dispatch.component import Component
from ribosome.trans.message_base import default_prio, Message

C = TypeVar('C', bound=Component)
D = TypeVar('D')
M = TypeVar('M', bound=Message)
R = TypeVar('R')
dp = default_prio
MDecorator = Callable[[Callable[[C], R]], MessageTransHandler[M, D]]
FDecorator = Callable[[Callable[[C], R]], FreeTransHandler[M, D]]


class MessageTransCons:

    def cons(self, msg_type: Type[M], *effects: TransEffect, prio: float=dp) -> MDecorator:
        def add_handler(func: Callable[[C], R]) -> MessageTransHandler[M, D]:
            return MessageTransHandler.create(func, msg_type, effects, prio)
        return add_handler

    def unit(self, msg_type: Type[M], *effects: TransEffect, prio: float=default_prio) -> MDecorator:
        return self.cons(msg_type, *effects, TransEffectUnit(), prio=prio)

    def one(self, tpe: Type[M], *effects: TransEffect, prio: float=dp) -> MDecorator:
        return self.cons(tpe, *effects, TransEffectSingleMessage(), prio=prio)

    def multi(self, tpe: Type[M], *effects: TransEffect, prio: float=dp) -> MDecorator:
        return self.cons(tpe, *effects, TransEffectMessages(), prio=prio)

    def relay(self, tpe: Type[M], prio: float=dp) -> Callable[[Callable[[C], R]], MessageTransHandler[M, D]]:
        def add_handler(func: Callable[[C], R]):
            return func
            # return decorate(func, tpe, prio)
        return add_handler


class FreeTransCons:

    def cons(self, *effects: TransEffect, prio: float=dp) -> FDecorator:
        def add_handler(func: Callable[[C], R]) -> FreeTransHandler[M, D]:
            return FreeTransHandler.create(func, Lists.wrap(effects), prio)
        return add_handler

    def unit(self, *effects: TransEffect, prio: float=default_prio) -> FDecorator:
        return self.cons(*effects, TransEffectUnit(), prio=prio)

    def one(self, *effects: TransEffect, prio: float=dp) -> FDecorator:
        return self.cons(*effects, TransEffectSingleMessage(), prio=prio)

    def multi(self, *effects: TransEffect, prio: float=dp) -> FDecorator:
        return self.cons(*effects, TransEffectMessages(), prio=prio)

    def result(self, *effects: TransEffect, prio: float=dp) -> FDecorator:
        return self.cons(*effects, TransEffectResult(), prio=prio)

    def do(self, prio: float=dp) -> FDecorator:
        return self.cons(TransEffectDo(), prio=prio)

__all__ = ('MessageTransCons', 'FreeTransCons')



import inspect
from typing import Callable, Any

from amino import Maybe, _, Just, Boolean, Lists, Nothing
from amino.dat import Dat

from ribosome.request.nargs import Nargs


class ParamsSpec(Dat['ParamsSpec']):

    @staticmethod
    def from_function(fun: Callable[..., Any]) -> 'ParamsSpec':
        argspec = inspect.getfullargspec(fun)
        params = Lists.wrap(argspec.args)
        defaults = Lists.wrap(argspec.defaults or ())
        method = Boolean(params.head.contains('self'))
        param_count = params.length - method.to_int
        min = param_count - defaults.length
        max = (~Boolean(argspec.varargs or argspec.varkw)).m(param_count)
        nargs = Nargs.cons(min, max)
        return ParamsSpec(nargs, min, max, method)

    def __init__(self, nargs: Nargs, min: int, max: Maybe[int], method: Boolean) -> None:
        self.nargs = nargs
        self.min = min
        self.max = max
        self.method = method

    @property
    def exact_count(self) -> Maybe[int]:
        return Just(self.min) if self.max.contains(self.min) else Nothing


class ArgValidator(Dat['ArgValidator']):

    def __init__(self, spec: ParamsSpec) -> None:
        self.spec = spec

    @property
    def min(self) -> int:
        return self.spec.min

    @property
    def max(self) -> Maybe[int]:
        return self.spec.max

    def validate(self, args: tuple) -> Boolean:
        l = len(args)
        return self.min <= l and not self.max.exists(_ < l)

    def error(self, args: tuple, desc: str, vim_name: str) -> str:
        return f'argument count for {desc} `{vim_name}` is {len(args)}, must be {self.count_spec} ({args})'

    @property
    def count_spec(self) -> str:
        return (
            self.spec.exact_count /
            (lambda a: f'exactly {a}' if a > 0 else 'none') | (
                self.max /
                (lambda mx: f'between {self.min} and {mx}') |
                f'at least {self.min}'
            )
        )

__all__ = ('ArgValidator', 'ParamsSpec')



import abc
from typing import Callable, TypeVar, Type, Generic, Awaitable, Any

from amino import Either, List, IO, L, _, Maybe, __
from amino.state import StateT
from amino.util.string import red, green
from amino.dispatch import dispatch_alg
from amino.dat import Dat
from amino.tc.functor import Functor
from amino.tc.base import Implicits

from ribosome.logging import Logging
from ribosome.nvim import NvimIO
from ribosome.trans.action import (TransStep, TransEffectError, Lift, Propagate, Strict, TransUnit, TransResult,
                                   TransAction, Transit, TransFailure, TransIO, TransM, TransDo)
from ribosome.trans.message_base import Sendable, Messages
from ribosome.trans.messages import Error, Nop, CoroutineAlg
from ribosome.process import Subprocess, SubprocessResult

D = TypeVar('D')
A = TypeVar('A')
B = TypeVar('B')
R = TypeVar('R')
N = TypeVar('N')
G = TypeVar('G')


class TransEffect(Generic[R], abc.ABC, Logging):

    @abc.abstractproperty
    def tpe(self) -> Type[R]:
        ...

    @abc.abstractmethod
    def extract(self, data: R, tail: List['TransEffect'], in_state: bool) -> TransStep:
        ...

    def typecheck(self, data: R) -> bool:
        return isinstance(data, self.tpe)

    def run(self, data: R, tail: List['TransEffect'], in_state: bool) -> TransStep:
        return (
            self.extract(data, tail, in_state)
            if self.typecheck(data) else
            TransEffectError(f'result {red(data)} does not have type {green(self.tpe.__qualname__)}')
        )


def cont(tail: List[TransEffect], in_state: bool, f: Callable[[Callable[[R], TransStep]], TransStep]
         ) -> Maybe[TransStep]:
    return tail.detach_head.map2(lambda h, t: f(lambda a: h.run(a, t, in_state)))


class TransEffectMaybe(Generic[R], TransEffect[Maybe[R]]):

    @property
    def tpe(self) -> Type[Maybe[R]]:
        return Maybe

    def extract(self, data: Maybe[N], tail: List[TransEffect], in_state: bool) -> TransStep:
        nested = data.map(lambda a: cont(tail, in_state, lambda run: Lift(run(a))) | a)
        return Lift(nested | Propagate.one(Nop()))


# FIXME use TransEffectError, not Error?
class TransEffectEither(Generic[R], TransEffect[Either[str, R]]):

    @property
    def tpe(self) -> Type[Either[str, R]]:
        return Either

    def extract(self, data: Either[str, R], tail: List[TransEffect], in_state: bool) -> TransStep:
        nested = data.map(lambda a: cont(tail, in_state, lambda run: Lift(run(a))) | a)
        return Lift(nested.value_or(L(Error)(_) >> Propagate.one))


class TransEffectStateT(Generic[G, D, R], TransEffect[StateT[G, D, R]]):

    def check(self, result: TransEffect[R]) -> None:
        return (
            TransEffectError(f'`State` trans result `{result}` is not a `TransStep`')
            if not isinstance(result, TransStep) else
            TransEffectError('found nested `State` in transition result')
            if result.strict else
            TransEffectError(result.data)
            if result.error else
            result.data
        )

    @property
    def tpe(self) -> Type[StateT[G, D, R]]:
        return StateT

    def extract(self, data: StateT[G, D, R], tail: List[TransEffect], in_state: bool) -> TransStep:
        return (
            TransEffectError('cannot nest `StateT` in transition result')
            if in_state else
            Strict(cont(tail, True, lambda run: data.map(lambda inner: self.check(run(inner)))) | data)
        )


class TransEffectIO(Generic[R], TransEffect[IO[R]]):

    @property
    def tpe(self) -> Type[IO[R]]:
        return IO

    def extract(self, data: IO[R], tail: List[TransEffect], in_state: bool) -> TransStep:
        io = cont(tail, False, data.map) | data
        return Lift(TransIO(io.map(L(lift)(_, in_state))))


class GatherIOs(Generic[A], Dat['GatherIOs'], Implicits, implicits=True, auto=True):

    def __init__(self, ios: List[IO[A]], handle_result: Callable[[List[A]], TransAction], timeout: float=None
                 ) -> None:
        self.ios = ios
        self.handle_result = handle_result
        self.timeout = timeout


class GatherIOsFunctor(Functor, tpe=GatherIOs):

    def map(self, fa: GatherIOs[A], f: Callable[[A], B]) -> None:
        return GatherIOs(fa.ios.map(__.map(f)), fa.handle_result, fa.timeout)


class TransEffectGatherIOs(TransEffect[GatherIOs]):

    @property
    def tpe(self) -> Type[GatherIOs]:
        return GatherIOs

    def extract(self, data: GatherIOs, tail: List[TransEffect], in_state: bool) -> TransStep:
        def handle_result(a: List[A]) -> TransAction:
            handled = data.handle_result(a)
            result = cont(tail, False, lambda f: f(handled)) | handled
            return lift(result, in_state)
        return Lift(TransIO(data.copy(handle_result=handle_result)))


class GatherSubprocs(Generic[A, R], Dat['GatherSubprocs'], Implicits, implicits=True, auto=True):

    def __init__(self, procs: List[Subprocess[A]], handle_result: Callable[[SubprocessResult[A]], R],
                 timeout: float=None) -> None:
        self.procs = procs
        self.handle_result = handle_result
        self.timeout = timeout


class TransEffectGatherSubprocs(TransEffect[GatherSubprocs]):

    @property
    def tpe(self) -> Type[GatherSubprocs]:
        return GatherSubprocs

    def extract(self, data: GatherSubprocs, tail: List[TransEffect], in_state: bool) -> TransStep:
        def handle_result(a: List[A]) -> TransAction:
            handled = data.handle_result(a)
            result = cont(tail, False, lambda f: f(handled)) | handled
            return lift(result, in_state)
        return Lift(TransIO(data.copy(handle_result=handle_result)))


class TransEffectNvimIO(Generic[R], TransEffect[NvimIO[R]]):

    @property
    def tpe(self) -> Type[NvimIO[R]]:
        return NvimIO

    def extract(self, data: NvimIO[R], tail: List[TransEffect], in_state: bool) -> TransStep:
        io = cont(tail, False, data.map) | data
        return Lift(TransIO(io.map(L(lift)(_, in_state))))


class TransEffectCoro(TransEffect):

    @property
    def tpe(self) -> Type[Awaitable]:
        return Awaitable

    def extract(self, data: Awaitable, tail: List[TransEffect], in_state: bool) -> Either[R, N]:
        async def coro_map(run: Callable[[R], TransStep]) -> TransStep:
            res = await data
            return lift(run(res), in_state)
        coro = cont(tail, False, coro_map) | data
        return Lift(Propagate.one(CoroutineAlg(coro).pub))


class TransEffectSingleMessage(TransEffect[Sendable]):

    @property
    def tpe(self) -> Type[Sendable]:
        return Messages

    def extract(self, data: Sendable, tail: List[TransEffect], in_state: bool) -> Either[R, N]:
        return Lift(Propagate.one(data)) if tail.empty else TransEffectError('cannot apply trans effects to Sendable')


class TransEffectMessages(TransEffect[List[Sendable]]):

    @property
    def tpe(self) -> Type[List[Sendable]]:
        return List

    def typecheck(self, data: List[Sendable]) -> bool:
        return super().typecheck(data) and data.forall(lambda a: isinstance(a, Messages))

    def extract(self, data: List[Sendable], tail: List[TransEffect], in_state: bool) -> Either[R, N]:
        return Lift(Propagate(data)) if tail.empty else TransEffectError('cannot apply trans effects to Messages')


class TransEffectUnit(TransEffect[None]):

    @property
    def tpe(self) -> Type[Sendable]:
        return type(None)

    def extract(self, data: None, tail: List[TransEffect], in_state: bool) -> Either[R, N]:
        return Lift(TransUnit()) if tail.empty else TransEffectError('cannot apply trans effects to unit')


class TransEffectResult(TransEffect[Any]):

    @property
    def tpe(self) -> Type[Any]:
        return object

    def extract(self, data: object, tail: List[TransEffect], in_state: bool) -> Either[R, N]:
        return Lift(TransResult(data)) if tail.empty else TransEffectError('cannot apply trans effects to result')


class TransEffectDo(TransEffect[TransM]):

    @property
    def tpe(self) -> Type[TransM]:
        return TransM

    def extract(self, data: TransM[A], tail: List[TransEffect], in_state: bool) -> Either[R, N]:
        return Lift(TransDo(data)) if tail.empty else TransEffectError('cannot apply trans effects to TransM[A]')


class Lifter(Logging):

    def lift_lift(self, res: Lift, in_state: bool) -> TransAction:
        return lift(res.data, in_state)

    def lift_strict(self, res: Strict, in_state: bool) -> TransAction:
        return Transit(res.data / L(lift)(_, True))

    def lift_trans_effect_error(self, res: TransEffectError, in_state: bool) -> TransAction:
        return TransFailure(res.data)

    def lift_res(self, res: R, in_state: bool) -> TransAction:
        return (
            res
            if isinstance(res, TransAction) else
            TransFailure(f'transition did not produce `TransAction`: {red(res)}')
        )


_lifter = Lifter()
lift = dispatch_alg(_lifter, TransStep, 'lift_', _lifter.lift_res)

__all__ = ('TransEffect', 'lift')



from typing import Generic, TypeVar, Tuple

from amino import List, Nil, Lists, Boolean
from amino.dat import Dat, DatMeta

A = TypeVar('A')

Item = Tuple[float, A]
Q = List[Item]


def insert(items: Q, a: A, prio: float) -> Q:
    item = (prio, a)
    index = items.index_where(lambda i: i[0] > prio) | -1
    pre, post = items.take(index), items.drop(index)
    return pre.cat(item) + post


class PrioQueueMeta(DatMeta):

    @property
    def empty(self) -> 'PrioQueue[A]':
        return PrioQueue(Nil)


class PrioQueue(Generic[A], Dat['PrioQueue[A]'], metaclass=PrioQueueMeta):

    def __init__(self, items: Q) -> None:
        self.items = items

    def put(self, a: A, prio: float) -> 'PrioQueue[A]':
        return PrioQueue(Lists.wrap(insert(self.items, a, prio)))

    def put_default(self, a: A) -> 'PrioQueue[A]':
        return self.put(a, 0.5)

    @property
    def get(self) -> 'Either[str, Tuple[PrioQueue[A], A]]':
        return self.items.detach_head.to_either('queue is empty').map2(lambda a, r: (a, PrioQueue(r)))

    @property
    def empty(self) -> Boolean:
        return self.items.empty

__all__ = ('PrioQueue',)



from kallikrein import Expectation, k
from kallikrein.matchers.either import be_right

from ribosome.test.integration.klk import AutoPluginIntegrationKlkSpec

from amino import List, __

from integration._support.envelope import P1, P2, T2, T3, S1, S2, S3


class EnvelopeSpec(AutoPluginIntegrationKlkSpec):
    '''
    publish a message $publish
    send message to specific machine $target
    process messages according to their priority values $prio
    '''

    def module(self) -> str:
        return 'integration._support.envelope'

    @property
    def plugin_prefix(self) -> str:
        return 'envl'

    def publish(self) -> Expectation:
        self.vim.cmd_once_defined('EnvlStage1')
        self.cmd_sync('EnvlPub')
        self.seen_message(P1)
        return self.seen_message(P2)

    def target(self) -> Expectation:
        self.vim.cmd_once_defined('EnvlStage1')
        self.cmd_sync('EnvlTarget')
        self.seen_message(T3)
        return self.seen_times(T2, 0)

    def prio(self) -> Expectation:
        self.vim.cmd_once_defined('EnvlStage1')
        self.cmd_sync('EnvlSeq')
        self.seen_message(S3)
        return k(self.message_log() / __.drop(2)).must(be_right(List(S2(), S3(), S1())))

__all__ = ('EnvelopeSpec',)



import abc
import functools
import time
import inspect
from typing import Optional, Any, TypeVar, Type, Generic
from types import FunctionType, SimpleNamespace

from amino import Map, Just, L, Maybe, _, Lists, Nothing
from amino.dat import Dat, DatMeta
from amino.util.ast import synth_init

_machine_attr = '_machine'
_message_attr = '_message'
_prio_attr = '_prio'
_dyn_attr = '_dyn'
default_prio = 0.5
fallback_prio = 0.3
override_prio = 0.8


class Sendable:

    @abc.abstractproperty
    def msg(self) -> 'Message':
        ...


M = TypeVar('M', bound='Message')


class Message(Generic[M], Dat[M], Sendable):

    def at(self, prio: float) -> 'Envelope[M]':
        return Envelope(self, time.time(), prio, Nothing)

    @property
    def envelope(self) -> 'Envelope[M]':
        return self.at(0.5)

    @property
    def pub(self) -> 'Envelope[M]':
        return self.envelope

    def to(self, target: str) -> 'Envelope[M]':
        return self.envelope.copy(recipient=Just(target))

    @property
    def msg(self) -> 'Message':
        return self


def message_init(fields: Map[str, Type], _globals: dict) -> FunctionType:
    params = fields.to_list
    return synth_init(params, _globals)


class MsgMeta(DatMeta):

    @classmethod
    def __prepare__(cls, _name: str, bases: tuple, glob: dict=None, **kw: Any) -> dict:
        globs = glob or inspect.currentframe().f_back.f_globals
        return dict(__init__=message_init(Map(kw), globs)) if kw else dict()

    def __new__(cls, _name: str, bases: tuple, namespace: SimpleNamespace, **kw: Any) -> type:
        return super().__new__(cls, _name, bases, namespace)


class Msg(Generic[M], Message[M], metaclass=MsgMeta):
    pass


def message_definition_module() -> Optional[str]:
    return inspect.currentframe().f_back.f_back.f_globals['__name__']


def pmessage(_name: str, *fields: str, mod: str=None, **kw: Any) -> Type[M]:
    module = mod or message_definition_module()
    params = Map(Lists.wrap(fields).apzip(lambda a: Any))
    ns = MsgMeta.__prepare__(_name, (Msg,), glob=dict(__module__=module), **params, **kw)
    return MsgMeta(_name, (Msg,), ns, **kw)


def json_pmessage(name, *fields, mod=None, **kw):
    module = mod or message_definition_module()
    return pmessage(name, *fields, mod=module, options=Map, **kw)


@functools.total_ordering
class Envelope(Generic[M], Dat['Envelope[M]'], Sendable):

    @staticmethod
    def from_sendable(s: Sendable) -> 'Envelope[M]':
        return s if isinstance(s, Envelope) else Envelope.default(s)

    @staticmethod
    def default(m: Message[M]) -> 'Envelope[M]':
        return Envelope(m, time.time(), 0.5, Nothing)

    def __init__(self, message: Message[M], time: float, prio: float, recipient: Maybe[str]) -> None:
        self.message = message
        self.time = time
        self.prio = prio
        self.recipient = recipient

    def at(self, prio: float) -> 'Envelope[M]':
        return self.copy(prio=prio)

    def __lt__(self, other):
        return (
            not isinstance(other, Envelope) or
            (
                self.time < self.time
                if self.prio == other.prio else
                self.prio < other.prio
            )
        )

    @property
    def delivery(self) -> Message:
        return self.recipient / L(ToMachine)(self, _) | self.message

    @property
    def msg(self) -> 'Message':
        return self.message


class ToMachine(Generic[M], Dat['ToMachine[M]'], Sendable):

    def __init__(self, envelope: Envelope[M], target: str) -> None:
        self.envelope = envelope
        self.target = target

    @property
    def message(self) -> Message[M]:
        return self.envelope.message

    @property
    def msg(self) -> Message[M]:
        return self.message


class Messages(abc.ABC):
    pass


Messages.register(Message)
Messages.register(Envelope)

__all__ = ('pmessage', 'json_pmessage', 'Sendable', 'Message', 'Msg', 'Envelope', 'ToMachine', 'Messages')



from amino.logging import LogError

from ribosome.trans.message_base import pmessage, json_pmessage, Message

Nop = pmessage('Nop')
Stop = pmessage('Stop')
Quit = pmessage('Quit')
Done = pmessage('Done')
PlugCommand = pmessage('PlugCommand', 'plug', 'msg')
Info = pmessage('Info', 'message')
ShowLogInfo = pmessage('ShowLogInfo')
RunIOsParallel = pmessage('RunIOsParallel', 'ios')
RunCorosParallel = pmessage('RunCorosParallel', 'coros')
SubProcessSync = pmessage('SubProcessSync', 'job', 'result')
SubProcessAsync = pmessage('SubProcessAsync', 'job', 'result')
Fork = pmessage('Fork', 'callback')
Warning = pmessage('Warning', 'message')
Debug = pmessage('Debug', 'message')
Coroutine = pmessage('Coroutine', 'coro')
CoroutineAlg = pmessage('CoroutineAlg', 'coro')
TransitionException = pmessage('TransitionException', 'context', 'exc')
UpdateRecord = json_pmessage('UpdateRecord', 'tpe', 'name')
UpdateState = json_pmessage('UpdateState', 'tpe', 'name')
Stage1 = pmessage('Stage1')
Stage2 = pmessage('Stage2')
Stage3 = pmessage('Stage3')
Stage4 = pmessage('Stage4')


class Error(Message['Error'], LogError):

    def __init__(self, main, prefix='') -> None:
        self.main = main
        self.prefix = prefix

    @property
    def message(self) -> str:
        pre = f'{self.prefix}: ' if self.prefix else ''
        return f'{pre}{self.main}'

    @property
    def full(self) -> str:
        return self.message

    @property
    def short(self) -> str:
        return self.main

__all__ = ('Nop', 'Stop', 'Quit', 'Done', 'Done', 'PlugCommand', 'Info', 'ShowLogInfo', 'RunIOsParallel',
           'RunCorosParallel', 'SubProcessSync', 'SubProcessAsync', 'Fork', 'Error', 'Warning', 'Debug', 'Coroutine',
           'CoroutineAlg', 'TransitionException', 'UpdateRecord', 'UpdateState', 'Stage1', 'Stage2', 'Stage3', 'Stage4')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x9b5f7ea0

# Compiled with Coconut version 1.3.0 [Dead Parrot]

# Coconut Header: -------------------------------------------------------------

from __future__ import generator_stop
import sys as _coconut_sys, os.path as _coconut_os_path
_coconut_file_path = _coconut_os_path.dirname(_coconut_os_path.abspath(__file__))
_coconut_sys.path.insert(0, _coconut_file_path)
from __coconut__ import _coconut, _coconut_NamedTuple, _coconut_MatchError, _coconut_tail_call, _coconut_tco, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_pipe, _coconut_star_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial
from __coconut__ import *
_coconut_sys.path.remove(_coconut_file_path)

# Compiled Coconut: -----------------------------------------------------------

from typing import TypeVar

from lenses import Lens

from amino.do import do
from amino import IO
from amino import List
from amino import Path
from amino import Left
from amino.json import decode_json
from amino.json import dump_json

from ribosome.nvim import NvimIO
from ribosome.nvim.io import NvimIOState
from ribosome.config import Data

A = TypeVar('A')
D = TypeVar('D', bound=Data)


@do(NvimIOState[D, Path])
def state_file(name: 'str') -> 'Generator':
    settings = yield NvimIOState.inspect(lambda a: a.settings)
    dir = yield NvimIOState.lift(settings.project_state_dir.value_or_default)
    yield NvimIOState.lift(NvimIO.from_io(IO.delay(dir.mkdir, parents=True, exist_ok=True)))
    yield NvimIOState.pure(dir / f'{name}.json')


@do(NvimIOState[D, None])
def load_json_data(name: 'str') -> 'Generator':
    file = yield state_file(name)
    exists = yield NvimIOState.lift(NvimIO.from_io(IO.delay(file.exists)))
    if exists:
        json = yield NvimIOState.lift(NvimIO.from_io(IO.delay(file.read_text)))
        yield NvimIOState.pure(decode_json(json))
    else:
        yield NvimIOState.pure(Left(f'state file {file} does not exist'))


@_coconut_tco
def load_json_state(name: 'str', l: 'Lens') -> 'NvimIOState[D, None]':
    return _coconut_tail_call(load_json_data(name).cata, lambda a: NvimIOState.pure(None), lambda d: NvimIOState.modify(lambda a: l.bind(a).set(d)))


@do(NvimIOState[D, None])
def store_json_data(name: 'str', data: 'A') -> 'Generator':
    file = yield state_file(name)
    json = yield NvimIOState.lift(NvimIO.from_either(dump_json(data)))
    yield NvimIOState.lift(NvimIO.from_io(IO.delay(file.write_text, json)))
    yield NvimIOState.pure(None)


@do(NvimIOState[D, None])
def store_json_state(name: 'str', data: '_coconut.typing.Callable[[D], A]') -> 'Generator':
    payload = yield NvimIOState.inspect(data)
    yield store_json_data(name, payload)

__all__ = ('load_json_state', 'store_json_data', 'store_json_state')



import inspect
from typing import Callable, Iterable, cast, Generator, TypeVar

from amino.options import env_xdg_data_dir
from amino import List, Either, Lists, Map, _, L, __, Path, Right, Try, Nil, Boolean, Left, Eval
from amino.func import flip
from amino.util.string import ToStr
from amino.do import do

from ribosome.nvim import NvimIO
from ribosome.logging import Logging
from ribosome.config.setting import PluginSetting, EvalSetting, setting_ctor

A = TypeVar('A')
B = TypeVar('B')
state_dir_help_default = '''This directory is used to persist the plugin's current state.'''
state_dir_base_default = env_xdg_data_dir.value / Path | (Path.home() / '.local' / 'share')
proteome_name_help = 'If **proteome** is installed, the session name is obtained from the main project name.'
session_name_help = 'A custom session name for the state dir can be specified.'
components_help = '''The plugin can run an arbitrary set of sub-components that inherit the class `Component`.
They receive all messages the core of the plugin processes and have access to the plugin state.
They can define nvim request handlers with the decorator `trans.xyz`.
The entries of this setting can either be names of the builtin components or arbitrary python module paths that define
custom components.
'''


@do(Either[str, str])
def project_name_from_path() -> Generator:
    cwd = yield Try(Path.cwd)
    name = cwd.name
    yield Right(name) if name else Left('cwd broken')


@do(NvimIO[Either[str, Path]])
def state_dir_with_name(state_dir: PluginSetting[Path], proteome_name: PluginSetting[str],
                        session_name: PluginSetting[str]) -> Generator:
    base = yield state_dir.value_or_default
    pro_name = yield proteome_name.value
    sess_name = yield session_name.value
    path = pro_name.o(sess_name).o(project_name_from_path) / (lambda a: base / a)
    yield NvimIO.pure(path)


class PluginSettings(ToStr, Logging):

    def __init__(self, name: str, state_dir_help: str=state_dir_help_default) -> None:
        self.name = name
        self.components = list_setting('components', 'names or paths of active components', components_help, True,
                                       Right(Nil))
        self.state_dir = path_setting('state_dir', 'state persistence directory', state_dir_help, True,
                                      Right(state_dir_base_default / self.name))
        self.proteome_name = str_setting('proteome_main_name', 'project name from protoeome', proteome_name_help, False)
        self.session_name = str_setting('ribosome_session_name', 'project name from user var', session_name_help, False)
        self.project_state_dir = EvalSetting(
            'project_state_dir',
            Eval.always(lambda: state_dir_with_name(self.state_dir, self.proteome_name, self.session_name))
        )

    def all(self) -> Map[str, PluginSetting]:
        settings = inspect.getmembers(self, L(isinstance)(_, PluginSetting))
        return Map(Lists.wrap(settings).map(_[1]).apzip(_.name).map2(flip))

    def _arg_desc(self) -> List[str]:
        return List(str(self.all()))


def path_list(data: list) -> Either[str, List[Path]]:
    return Lists.wrap(data).traverse(lambda a: Try(Path, a) / __.expanduser(), Either)


str_setting = setting_ctor(str, lambda a: Right(a))
float_setting = setting_ctor(float, lambda a: Right(a))
list_setting = setting_ctor(list, cast(Callable[[Iterable[A]], Either[str, List[B]]], (lambda a: Right(Lists.wrap(a)))))
path_setting = setting_ctor(str, (lambda a: Try(Path, a)))
path_list_setting = setting_ctor(list, path_list)
map_setting = setting_ctor(dict, lambda a: Right(Map(a)))
path_map_setting = setting_ctor(dict, lambda a: Try(Map, a).valmap(lambda b: Path(b).expanduser()))
bool_setting = setting_ctor(int, lambda a: Right(Boolean(a)))


__all__ = ('PluginSettings', 'str_setting', 'float_setting', 'list_setting', 'path_setting', 'path_list_setting',
           'map_setting', 'path_map_setting', 'bool_setting')



import abc

from amino import List, Maybe, Just, Nothing
from amino.dat import ADT


class Nargs(ADT):

    @staticmethod
    def cons(min: int, mmax: Maybe[int]) -> 'Nargs':
        def from_min() -> Nargs:
            return NargsStar() if min == 0 else NargsPlus()
        def with_max(max: int) -> Nargs:
            return (
                Just(NargsZero())
                if max == 0 else
                (
                    Just(NargsOne())
                    if min == 1 else
                    Just(NargsQM())
                )
                if max == 1 else
                Nothing
            )
        return mmax.flat_map(with_max) | from_min

    @abc.abstractproperty
    def for_vim(self) -> str:
        ...

    def _arg_desc(self) -> List[str]:
        return List()


class NargsZero(Nargs):

    @property
    def for_vim(self) -> str:
        return '0'


class NargsOne(Nargs):

    @property
    def for_vim(self) -> str:
        return '1'


class NargsStar(Nargs):

    @property
    def for_vim(self) -> str:
        return '*'


class NargsPlus(Nargs):

    @property
    def for_vim(self) -> str:
        return '+'


class NargsQM(Nargs):

    @property
    def for_vim(self) -> str:
        return '?'

__all__ = ('Nargs',)



from ribosome import command, function, NvimStatePlugin

from ribosome.logging import Logging


class FacadeTestPlugin(NvimStatePlugin, Logging):

    def state(self) -> None:
        pass

    @function()
    def all_vars(self):
        return self.vim.buffer.vars.all

    @command(sync=True)
    def go(self):
        pass

__all__ = ('FacadeTestPlugin',)



import abc
import inspect
from numbers import Number
from typing import Callable, Any, Union

from amino import List, Maybe, Map
from amino.util.string import camelcase
from amino.boolean import true, false

from ribosome.logging import Logging
from ribosome.request.args import ParamsSpec, ArgValidator
from ribosome.request.handler.handler import RequestHandler
from ribosome.request.handler.dispatcher import RequestDispatcher, FunctionDispatcher
from ribosome.request.handler.prefix import Plain
from ribosome.request.handler.method import RpcMethod


def to_int(val: Union[int, str, None]) -> Union[int, str, None]:
    return int(val) if isinstance(val, str) and str.isdigit() else val


def numeric(val: Any) -> bool:
    return isinstance(val, Number)


class ParseError(Exception):
    pass


class LegacyRequestHandler(Logging, metaclass=abc.ABCMeta):

    def __init__(
            self,
            fun: Callable[..., Any],
            name: str=None,
            nargs: Union[str, int]=None,
            min: int=None,
            max: int=None,
            **kw: str
    ) -> None:
        self._fun = fun
        self._argspec = inspect.getfullargspec(fun)
        self._params = List.wrap(self._argspec.args)
        self._param_count = self._params.length - (1 if self._params.head.contains('self') else 0)
        self._name = Maybe(name)
        self._nargs = Maybe(to_int(nargs))
        self._min = Maybe(min)
        self._max = Maybe(max)
        self._kw = Map(kw)
        self.params_spec = ParamsSpec.from_function(self._fun)

    @abc.abstractproperty
    def desc(self):
        ...

    @property
    def kw(self):
        return self.default_kw ** self._kw

    @property
    def default_kw(self):
        return Map()

    @property
    def nargs(self) -> Union[str, int]:
        return self._nargs\
            .map(lambda a: '+' if numeric(a) and a > 1 else a)\
            .get_or_else(self.params_spec.nargs.for_vim)

    @property
    def min(self):
        return self._min.get_or_else(self.params_spec.min)

    @property
    def max(self):
        return self._max.or_else(self.params_spec.max)

    @property
    def vim_name(self) -> str:
        return camelcase(self.name)

    @property
    def name(self) -> str:
        return self._name.get_or_else(self._infer_name)

    @property
    def _infer_name(self):
        return self._fun.__name__

    def _call_fun(self, obj, *args, **kw):
        return self._fun(obj, *args)

    def dispatch(self, obj, rpc_args):
        argl = List.wrap(rpc_args)
        args = argl.head | List()
        bang = argl.lift(1).contains(1)
        validator = ArgValidator(self.params_spec)
        if validator.validate(args):
            return self._call_fun(obj, *args, bang=bang)
        else:
            return validator.error(args, self.desc, self.vim_name)

    @abc.abstractproperty
    def method(self) -> RpcMethod:
        ...

    @property
    def dispatcher(self) -> RequestDispatcher:
        return FunctionDispatcher(self._fun)

    @property
    def spec(self) -> RequestHandler:
        return RequestHandler(self.method, self.dispatcher, self.name, Plain(), false, true, self.kw)

__all__ = ('LegacyRequestHandler',)



import abc
from typing import Callable, TypeVar, Any, Generic, Type, Tuple, Union, Generator
from asyncio import iscoroutinefunction

from amino.tc.optional import Optional
from amino import Maybe, Either, Just, Left, I, List, Nothing, Id, Nil
from amino.state import StateT
from amino.util.string import ToStr
from amino.dat import Dat, DatMeta
from amino.do import do

from ribosome.trans.message_base import _message_attr, Message, default_prio, _prio_attr, _dyn_attr

from ribosome.logging import Logging
from ribosome.trans.messages import Error, Debug, Coroutine

A = TypeVar('A')


class Failure:

    def __init__(self, message: str) -> None:
        self.message = message

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.message)


class Fatal(Failure):
    pass


class NothingToDo(Failure):
    pass


def _to_error(data):
    return (
        data if isinstance(data, Message)
        else Error(data.message) if isinstance(data, Fatal)
        else Debug(data.message) if isinstance(data, NothingToDo)
        else Error(data) if isinstance(data, Exception)
        else Debug(str(data))
    )


def _recover_error(handler: Any, result: Any) -> Maybe[Any]:
    if not Optional.exists(type(result)):
        err = 'in {}: result has no Optional: {}'
        return Just(Error(err.format(handler, result)))
    elif isinstance(result, Either):
        return Just(result.right_or_map(_to_error))
    else:
        return result


def _io_result(result):
    return result.cata(Left, I)

D = TypeVar('D')
M = TypeVar('M', bound=Message)
R = TypeVar('R')
DynResult = Union[Message, List[Message]]
DynTrans = Union[DynResult, Maybe[DynResult], Either[str, DynResult], StateT[Id, D, DynResult]]


class Handler:
    pass


class LegacyHandler(Generic[M, D, R], Handler, Logging, ToStr):

    def __init__(self, name: str, fun: Callable[[D, M], R], message: Type[M], prio: float, dyn: bool) -> None:
        self.name = name
        self.message = message
        self.fun = fun
        self.prio = prio
        self.dyn = dyn

    @staticmethod
    def attrs(fun: Callable) -> Tuple[str, Type[M], float, bool]:
        name = fun.__name__
        msg = getattr(fun, _message_attr, None)
        prio = getattr(fun, _prio_attr, default_prio)
        dyn = getattr(fun, _dyn_attr, False)
        return name, msg, prio, dyn

    @staticmethod
    def create(fun: Callable[[D, M], R]) -> 'LegacyHandler[M, D, R]':
        name, msg, prio, dyn = LegacyHandler.attrs(fun)
        tpe = CoroHandler if iscoroutinefunction(fun) else DynHandler  # if dyn else TransHandler
        return tpe(name, fun, msg, prio, dyn)

    @abc.abstractmethod
    def execute(self, data: D, msg: M) -> R:
        ...

    @abc.abstractmethod
    def run(self, data: D, msg: M) -> R:
        ...

    def _arg_desc(self) -> List[str]:
        return List(self.name, str(self.message), str(self.fun))


class DynHandler(Generic[M, D], LegacyHandler[M, D, DynTrans]):

    def run(self, data, msg) -> DynTrans:
        return _recover_error(self, self.execute(data, msg))

    def execute(self, data, msg) -> DynTrans:
        return self.fun(data, msg)


class CoroHandler(LegacyHandler):

    def run(self, data, msg) -> Maybe[Coroutine]:
        return Maybe(Coroutine(self.fun(data, msg)))

    def execute(self, data, msg) -> Maybe[Coroutine]:
        ...


class CoroExecutionHandler(DynHandler):
    pass


class TransitionLogMeta(DatMeta):

    @property
    def empty(self) -> 'TransitionLog':
        return TransitionLog(Nil, Nil)


class TransitionLog(Dat['TransitionLog'], metaclass=TransitionLogMeta):

    def __init__(self, message_log: List[Message], message_resend: List[Message]) -> None:
        self.message_log = message_log
        self.message_resend = message_resend

    def log(self, m: Message) -> 'TransitionLog':
        return self.append1.message_log(m)

    def resend(self, ms: List[Message]) -> 'TransitionLog':
        return self.append.message_resend(ms)

    @property
    def pop(self) -> Generator:
        @do(Maybe[Tuple['TransitionLog', Maybe[Message]]])
        def pop1() -> Generator:
            head, tail = yield self.message_resend.detach_head
            yield Just((self.copy(message_resend=tail).log(head), Just(head)))
        return pop1() | (self, Nothing)


class MachineError(RuntimeError):
    pass


class TransitionFailed(MachineError):
    pass

__all__ = ('LegacyHandler', 'CoroHandler', 'handle', 'may_handle', 'either_msg', 'either_handle', 'MachineError', 'TransitionFailed', 'TransitionLog')



from setuptools import setup, find_packages

version_parts = (13, 0, 0, 'a', 2)
version = '.'.join(map(str, version_parts))

setup(
    name='ribosome',
    description='neovim plugin framework',
    version=version,
    author='Torsten Schmits',
    author_email='torstenschmits@gmail.com',
    license='MIT',
    url='https://github.com/tek/ribosome',
    packages=find_packages(exclude=['unit', 'unit.*']),
    install_requires=[
        'amino==12.5.3',
        'neovim~=0.1.13',
        'pyuv~=1.4.0',
    ],
    tests_require=[
        'libtmux~=0.7.4',
        'kallikrein~=0.21.0',
    ],
    entry_points={
        'console_scripts': [
            'ribosome_start_plugin = ribosome.cli:start_plugin',
        ],
    },
)



from amino import Map, __, List

from ribosome.plugin import Config
from ribosome.trans.message_base import pmessage
from ribosome.trans.api import trans
from ribosome.trans.messages import Stage1
from ribosome.config import RequestHandler
from ribosome.nvim import NvimIO
from ribosome.dispatch.component import Component
from ribosome.request.handler.prefix import Plain

Msg1 = pmessage('Msg1')
Msg2 = pmessage('Msg2')
val = 71


class SpecCore(Component):

    @trans.msg.unit(Stage1)
    def stage_1(self) -> None:
        pass

    @trans.msg.unit(Msg1, trans.nio)
    def msg1(self) -> NvimIO[None]:
        return NvimIO.delay(__.vars.set('msg_cmd_success', val))

    @trans.msg.unit(Msg2, trans.nio)
    def msg2(self) -> NvimIO[None]:
        return NvimIO.delay(__.vars.set('autocmd_success', val))


config = Config.cons(
    name='plug',
    components=Map(core=SpecCore),
    request_handlers=List(
        RequestHandler.msg_cmd(Msg1)('msg1', prefix=Plain()),
        RequestHandler.msg_autocmd(Msg2)('vim_resized', prefix=Plain()),
    ),
)

__all__ = ('config',)
